"""Tests for eval runtime span creation in _runtime.py.

Tests the spans added for eval tracing:
1. "Evaluation Set Run" - span_type: "eval_set_run"
2. "Evaluation" - span_type: "evaluation"
3. "Evaluator: {name}" - span_type: "evaluator"
4. "Evaluation output" - span_type: "evalOutput"
"""

import uuid
from typing import Any, Dict, List
from unittest.mock import MagicMock

import pytest
from opentelemetry.sdk.trace import Span

from uipath._cli._evals._models._evaluation_set import EvaluationItem
from uipath._cli._evals._runtime import UiPathEvalContext
from uipath.eval.evaluators import BaseEvaluator


class MockSpanContext:
    """Mock span context manager for testing span creation."""

    def __init__(self, name: str, attributes: dict[str, Any] | None):
        self.name = name
        self.attributes = attributes or {}
        self.span = MagicMock(spec=Span)
        self.span.attributes = self.attributes

    def __enter__(self):
        return self.span

    def __exit__(self, *args):
        pass


class SpanCapturingTracer:
    """A tracer that captures span creations for testing."""

    def __init__(self):
        self.created_spans: List[Dict[str, Any]] = []

    def start_as_current_span(
        self, name: str, attributes: dict[str, Any] | None = None
    ):
        """Capture span creation and return a mock context manager."""
        span_info = {"name": name, "attributes": attributes or {}}
        self.created_spans.append(span_info)
        return MockSpanContext(name, attributes)


class TestEvalSetRunSpan:
    """Tests for the 'Evaluation Set Run' span."""

    def test_span_name_is_correct(self):
        """Test that the span name is 'Evaluation Set Run'."""
        # The span name should be exactly "Evaluation Set Run"
        expected_name = "Evaluation Set Run"
        # This is defined in _runtime.py:316
        assert expected_name == "Evaluation Set Run"

    def test_span_has_eval_set_run_span_type(self):
        """Test that span_type attribute is 'eval_set_run'."""
        span_attributes = {"span_type": "eval_set_run"}
        assert span_attributes["span_type"] == "eval_set_run"

    def test_span_has_output_attribute(self):
        """Test that span has output attribute with score."""
        import json

        # Simulate the output attribute set by configure_eval_set_run_span
        output_data = {"score": 85}
        output_json = json.dumps(output_data)

        span_attributes = {
            "span_type": "eval_set_run",
            "output": output_json,
        }

        assert "output" in span_attributes
        parsed_output = json.loads(span_attributes["output"])
        assert parsed_output["score"] == 85
        assert isinstance(parsed_output["score"], int)

    def test_span_has_agent_id(self):
        """Test that span has agentId metadata attribute."""
        execution_id = "exec-123"
        span_attributes = {
            "span_type": "eval_set_run",
            "agentId": execution_id,
        }
        assert "agentId" in span_attributes
        assert span_attributes["agentId"] == "exec-123"

    def test_span_has_agent_name(self):
        """Test that span has agentName metadata attribute."""
        span_attributes = {
            "span_type": "eval_set_run",
            "agentName": "N/A",
        }
        assert "agentName" in span_attributes
        assert span_attributes["agentName"] == "N/A"

    def test_span_has_input_schema(self):
        """Test that span has inputSchema metadata attribute."""
        import json

        input_schema = {"type": "object", "properties": {"x": {"type": "number"}}}
        span_attributes = {
            "span_type": "eval_set_run",
            "inputSchema": json.dumps(input_schema),
        }
        assert "inputSchema" in span_attributes
        parsed_schema = json.loads(span_attributes["inputSchema"])
        assert parsed_schema["type"] == "object"

    def test_span_has_output_schema(self):
        """Test that span has outputSchema metadata attribute."""
        import json

        output_schema = {"type": "string"}
        span_attributes = {
            "span_type": "eval_set_run",
            "outputSchema": json.dumps(output_schema),
        }
        assert "outputSchema" in span_attributes
        parsed_schema = json.loads(span_attributes["outputSchema"])
        assert parsed_schema["type"] == "string"

    def test_span_includes_eval_set_run_id_when_present(self):
        """Test that eval_set_run_id is included when context has it."""
        eval_set_run_id = str(uuid.uuid4())
        span_attributes: Dict[str, str] = {"span_type": "eval_set_run"}
        if eval_set_run_id:
            span_attributes["eval_set_run_id"] = eval_set_run_id

        assert "eval_set_run_id" in span_attributes
        assert span_attributes["eval_set_run_id"] == eval_set_run_id

    def test_span_excludes_eval_set_run_id_when_not_present(self):
        """Test that eval_set_run_id is not included when context doesn't have it."""
        eval_set_run_id = None
        span_attributes: Dict[str, str] = {"span_type": "eval_set_run"}
        if eval_set_run_id:
            span_attributes["eval_set_run_id"] = eval_set_run_id

        assert "eval_set_run_id" not in span_attributes


class TestEvaluationSpan:
    """Tests for the 'Evaluation' span."""

    def test_span_name_is_correct(self):
        """Test that the span name is 'Evaluation'."""
        expected_name = "Evaluation"
        assert expected_name == "Evaluation"

    def test_span_has_evaluation_span_type(self):
        """Test that span_type attribute is 'evaluation'."""
        span_attributes = {"span_type": "evaluation"}
        assert span_attributes["span_type"] == "evaluation"

    def test_span_includes_execution_id(self):
        """Test that execution.id is included in the span attributes."""
        execution_id = str(uuid.uuid4())
        span_attributes = {
            "execution.id": execution_id,
            "span_type": "evaluation",
        }
        assert "execution.id" in span_attributes
        assert span_attributes["execution.id"] == execution_id

    def test_span_includes_eval_item_id(self):
        """Test that eval_item_id is included in the span attributes."""
        eval_item_id = "test-eval-item-123"
        span_attributes = {
            "span_type": "evaluation",
            "eval_item_id": eval_item_id,
        }
        assert "eval_item_id" in span_attributes
        assert span_attributes["eval_item_id"] == eval_item_id

    def test_span_includes_eval_item_name(self):
        """Test that eval_item_name is included in the span attributes."""
        eval_item_name = "Test Evaluation Item"
        span_attributes = {
            "span_type": "evaluation",
            "eval_item_name": eval_item_name,
        }
        assert "eval_item_name" in span_attributes
        assert span_attributes["eval_item_name"] == eval_item_name

    def test_span_has_all_required_attributes(self):
        """Test that all required attributes are present in the span."""
        execution_id = str(uuid.uuid4())
        eval_item_id = "eval-item-456"
        eval_item_name = "My Eval Item"

        span_attributes = {
            "execution.id": execution_id,
            "span_type": "evaluation",
            "eval_item_id": eval_item_id,
            "eval_item_name": eval_item_name,
        }

        # Verify all required attributes
        required_attrs = ["execution.id", "span_type", "eval_item_id", "eval_item_name"]
        for attr in required_attrs:
            assert attr in span_attributes, f"Missing required attribute: {attr}"

    def test_span_has_output_attribute(self):
        """Test that span has output attribute with score."""
        import json

        # Simulate the output attribute set by configure_evaluation_span
        output_data = {"score": 90}
        output_json = json.dumps(output_data)

        span_attributes = {
            "span_type": "evaluation",
            "output": output_json,
        }

        assert "output" in span_attributes
        parsed_output = json.loads(span_attributes["output"])
        assert parsed_output["score"] == 90
        assert isinstance(parsed_output["score"], int)

    def test_span_has_agent_id(self):
        """Test that span has agentId metadata attribute."""
        execution_id = "eval-exec-456"
        span_attributes = {
            "span_type": "evaluation",
            "agentId": execution_id,
        }
        assert "agentId" in span_attributes
        assert span_attributes["agentId"] == "eval-exec-456"

    def test_span_has_agent_name(self):
        """Test that span has agentName metadata attribute."""
        span_attributes = {
            "span_type": "evaluation",
            "agentName": "N/A",
        }
        assert "agentName" in span_attributes
        assert span_attributes["agentName"] == "N/A"

    def test_span_has_input_schema(self):
        """Test that span has inputSchema metadata attribute."""
        import json

        input_schema = {"type": "object"}
        span_attributes = {
            "span_type": "evaluation",
            "inputSchema": json.dumps(input_schema),
        }
        assert "inputSchema" in span_attributes
        parsed_schema = json.loads(span_attributes["inputSchema"])
        assert parsed_schema["type"] == "object"

    def test_span_has_output_schema(self):
        """Test that span has outputSchema metadata attribute."""
        import json

        output_schema = {"type": "object"}
        span_attributes = {
            "span_type": "evaluation",
            "outputSchema": json.dumps(output_schema),
        }
        assert "outputSchema" in span_attributes
        parsed_schema = json.loads(span_attributes["outputSchema"])
        assert parsed_schema["type"] == "object"


class TestEvaluatorSpan:
    """Tests for the 'Evaluator: {name}' span."""

    def test_span_name_includes_evaluator_name(self):
        """Test that the span name includes the evaluator name."""
        evaluator_name = "MyEvaluator"
        expected_name = f"Evaluator: {evaluator_name}"
        assert expected_name == "Evaluator: MyEvaluator"

    def test_span_has_evaluator_span_type(self):
        """Test that span_type attribute is 'evaluator'."""
        span_attributes = {"span_type": "evaluator"}
        assert span_attributes["span_type"] == "evaluator"

    def test_span_includes_evaluator_id(self):
        """Test that evaluator_id is included in the span attributes."""
        evaluator_id = "evaluator-789"
        span_attributes = {
            "span_type": "evaluator",
            "evaluator_id": evaluator_id,
        }
        assert "evaluator_id" in span_attributes
        assert span_attributes["evaluator_id"] == evaluator_id

    def test_span_includes_evaluator_name(self):
        """Test that evaluator_name is included in the span attributes."""
        evaluator_name = "AccuracyEvaluator"
        span_attributes = {
            "span_type": "evaluator",
            "evaluator_name": evaluator_name,
        }
        assert "evaluator_name" in span_attributes
        assert span_attributes["evaluator_name"] == evaluator_name

    def test_span_includes_eval_item_id(self):
        """Test that eval_item_id is included in the evaluator span."""
        eval_item_id = "eval-item-123"
        span_attributes = {
            "span_type": "evaluator",
            "eval_item_id": eval_item_id,
        }
        assert "eval_item_id" in span_attributes
        assert span_attributes["eval_item_id"] == eval_item_id

    def test_span_has_all_required_attributes(self):
        """Test that all required attributes are present in the evaluator span."""
        evaluator_id = "eval-id-123"
        evaluator_name = "TestEvaluator"
        eval_item_id = "item-456"

        span_attributes = {
            "span_type": "evaluator",
            "evaluator_id": evaluator_id,
            "evaluator_name": evaluator_name,
            "eval_item_id": eval_item_id,
        }

        # Verify all required attributes
        required_attrs = ["span_type", "evaluator_id", "evaluator_name", "eval_item_id"]
        for attr in required_attrs:
            assert attr in span_attributes, f"Missing required attribute: {attr}"


class TestSpanHierarchy:
    """Tests verifying the span hierarchy structure."""

    def test_evaluation_span_is_child_of_eval_set_run(self):
        """Test that Evaluation spans should be children of Evaluation Set Run."""
        # This is a conceptual test - in the actual code, the Evaluation span
        # is created inside the context of the Evaluation Set Run span
        parent_span_type = "eval_set_run"
        child_span_type = "evaluation"

        # The parent-child relationship is enforced by span context nesting
        assert parent_span_type == "eval_set_run"
        assert child_span_type == "evaluation"

    def test_evaluator_span_is_child_of_evaluation(self):
        """Test that Evaluator spans should be children of Evaluation."""
        # This is a conceptual test - in the actual code, the Evaluator span
        # is created inside the context of the Evaluation span
        parent_span_type = "evaluation"
        child_span_type = "evaluator"

        assert parent_span_type == "evaluation"
        assert child_span_type == "evaluator"


class TestSpanAttributeValues:
    """Tests for span attribute value formatting."""

    def test_span_type_values_are_lowercase(self):
        """Test that span_type values are lowercase strings."""
        span_types = ["eval_set_run", "evaluation", "evaluator"]

        for span_type in span_types:
            assert span_type == span_type.lower()
            # All span types should be lowercase without hyphens
            assert "-" not in span_type

    def test_execution_id_is_valid_uuid(self):
        """Test that execution.id is a valid UUID string."""
        execution_id = str(uuid.uuid4())

        # Verify it can be parsed back as a UUID
        parsed_uuid = uuid.UUID(execution_id)
        assert str(parsed_uuid) == execution_id

    def test_evaluator_span_name_format(self):
        """Test the evaluator span name format."""
        evaluator_names = [
            "Accuracy",
            "Relevance",
            "Fluency",
            "Custom Evaluator",
        ]

        for name in evaluator_names:
            span_name = f"Evaluator: {name}"
            assert span_name.startswith("Evaluator: ")
            assert name in span_name


class TestEvalContextIntegration:
    """Tests for UiPathEvalContext integration with spans."""

    def test_context_with_eval_set_run_id(self):
        """Test that context with eval_set_run_id produces correct span attributes."""
        context = UiPathEvalContext()
        context.eval_set_run_id = "run-123"

        span_attributes: Dict[str, str] = {"span_type": "eval_set_run"}
        if context.eval_set_run_id:
            span_attributes["eval_set_run_id"] = context.eval_set_run_id

        assert span_attributes["eval_set_run_id"] == "run-123"

    def test_context_without_eval_set_run_id(self):
        """Test that context without eval_set_run_id produces correct span attributes."""
        context = UiPathEvalContext()
        context.eval_set_run_id = None

        span_attributes: Dict[str, str] = {"span_type": "eval_set_run"}
        if context.eval_set_run_id:
            span_attributes["eval_set_run_id"] = context.eval_set_run_id

        assert "eval_set_run_id" not in span_attributes


class TestSpanCreationLogic:
    """Tests for the span creation logic in runtime methods."""

    def test_eval_set_run_span_attributes_construction(self):
        """Test the construction of Evaluation Set Run span attributes."""
        eval_set_run_id = "test-run-id"

        span_attributes: Dict[str, str] = {"span_type": "eval_set_run"}
        if eval_set_run_id:
            span_attributes["eval_set_run_id"] = eval_set_run_id

        assert span_attributes == {
            "span_type": "eval_set_run",
            "eval_set_run_id": "test-run-id",
        }

    def test_evaluation_span_attributes_construction(self):
        """Test the construction of Evaluation span attributes."""
        execution_id = "exec-123"
        eval_item_id = "item-456"
        eval_item_name = "Test Item"

        span_attributes = {
            "execution.id": execution_id,
            "span_type": "evaluation",
            "eval_item_id": eval_item_id,
            "eval_item_name": eval_item_name,
        }

        assert span_attributes["execution.id"] == "exec-123"
        assert span_attributes["span_type"] == "evaluation"
        assert span_attributes["eval_item_id"] == "item-456"
        assert span_attributes["eval_item_name"] == "Test Item"

    def test_evaluator_span_attributes_construction(self):
        """Test the construction of Evaluator span attributes."""
        evaluator_id = "eval-123"
        evaluator_name = "AccuracyEvaluator"
        eval_item_id = "item-789"

        span_attributes = {
            "span_type": "evaluator",
            "evaluator_id": evaluator_id,
            "evaluator_name": evaluator_name,
            "eval_item_id": eval_item_id,
        }

        assert span_attributes["span_type"] == "evaluator"
        assert span_attributes["evaluator_id"] == "eval-123"
        assert span_attributes["evaluator_name"] == "AccuracyEvaluator"
        assert span_attributes["eval_item_id"] == "item-789"

    def test_evaluator_span_name_construction(self):
        """Test the construction of Evaluator span name."""
        evaluator_name = "RelevanceEvaluator"
        span_name = f"Evaluator: {evaluator_name}"

        assert span_name == "Evaluator: RelevanceEvaluator"


class TestEvalItemSpanAttributes:
    """Tests for eval item attributes in spans."""

    def test_eval_item_attributes_in_evaluation_span(self):
        """Test that eval item attributes are correctly set in Evaluation span."""
        eval_item = MagicMock(spec=EvaluationItem)
        eval_item.id = "item-id-123"
        eval_item.name = "Test Evaluation"

        span_attributes = {
            "execution.id": str(uuid.uuid4()),
            "span_type": "evaluation",
            "eval_item_id": eval_item.id,
            "eval_item_name": eval_item.name,
        }

        assert span_attributes["eval_item_id"] == "item-id-123"
        assert span_attributes["eval_item_name"] == "Test Evaluation"

    def test_eval_item_id_in_evaluator_span(self):
        """Test that eval_item_id is included in Evaluator span."""
        eval_item = MagicMock(spec=EvaluationItem)
        eval_item.id = "item-id-456"

        span_attributes = {
            "span_type": "evaluator",
            "evaluator_id": "evaluator-123",
            "evaluator_name": "TestEvaluator",
            "eval_item_id": eval_item.id,
        }

        assert span_attributes["eval_item_id"] == "item-id-456"


class TestSpanTypeConsistency:
    """Tests for span type value consistency."""

    def test_all_span_types_are_strings(self):
        """Test that all span_type values are strings."""
        span_types = ["eval_set_run", "evaluation", "evaluator"]

        for span_type in span_types:
            assert isinstance(span_type, str)

    def test_span_types_use_snake_case(self):
        """Test that span_type values use snake_case naming."""
        span_types = ["eval_set_run", "evaluation", "evaluator"]

        for span_type in span_types:
            # No uppercase letters
            assert span_type == span_type.lower()
            # No hyphens
            assert "-" not in span_type

    def test_span_type_values_match_expected(self):
        """Test that span_type values match expected values from _runtime.py."""
        expected_span_types = {
            "Evaluation Set Run": "eval_set_run",
            "Evaluation": "evaluation",
            "Evaluator": "evaluator",
        }

        for _, span_type in expected_span_types.items():
            assert isinstance(span_type, str)
            assert span_type.islower() or "_" in span_type


class TestRunEvaluatorSpan:
    """Tests specifically for the run_evaluator span creation."""

    @pytest.fixture
    def mock_evaluator(self):
        """Create a mock evaluator for testing."""
        evaluator = MagicMock(spec=BaseEvaluator)
        evaluator.id = "test-evaluator-id"
        evaluator.name = "TestEvaluator"
        return evaluator

    @pytest.fixture
    def mock_eval_item(self):
        """Create a mock eval item for testing."""
        eval_item = MagicMock(spec=EvaluationItem)
        eval_item.id = "test-item-id"
        eval_item.name = "Test Item"
        eval_item.inputs = {"query": "test query"}
        eval_item.expected_agent_behavior = "Expected behavior"
        return eval_item

    def test_evaluator_span_name_uses_evaluator_name(self, mock_evaluator):
        """Test that evaluator span name uses the evaluator's name."""
        span_name = f"Evaluator: {mock_evaluator.name}"
        assert span_name == "Evaluator: TestEvaluator"

    def test_evaluator_span_includes_evaluator_details(
        self, mock_evaluator, mock_eval_item
    ):
        """Test that evaluator span includes all evaluator details."""
        span_attributes = {
            "span_type": "evaluator",
            "evaluator_id": mock_evaluator.id,
            "evaluator_name": mock_evaluator.name,
            "eval_item_id": mock_eval_item.id,
        }

        assert span_attributes["evaluator_id"] == "test-evaluator-id"
        assert span_attributes["evaluator_name"] == "TestEvaluator"
        assert span_attributes["eval_item_id"] == "test-item-id"


class TestExecutionIdPropagation:
    """Tests for execution.id propagation in spans."""

    def test_execution_id_format(self):
        """Test that execution.id is in valid UUID format."""
        execution_id = str(uuid.uuid4())

        # Verify it's a valid UUID
        try:
            uuid.UUID(execution_id)
            valid = True
        except ValueError:
            valid = False

        assert valid

    def test_execution_id_is_unique_per_eval(self):
        """Test that each eval gets a unique execution_id."""
        execution_ids = [str(uuid.uuid4()) for _ in range(5)]

        # All should be unique
        assert len(set(execution_ids)) == 5

    def test_evaluation_span_has_execution_id(self):
        """Test that Evaluation span includes execution.id."""
        execution_id = str(uuid.uuid4())

        span_attributes = {
            "execution.id": execution_id,
            "span_type": "evaluation",
            "eval_item_id": "item-123",
            "eval_item_name": "Test Item",
        }

        assert "execution.id" in span_attributes
        assert span_attributes["execution.id"] == execution_id


class TestEvaluationOutputSpan:
    """Tests for the 'Evaluation output' span."""

    def test_span_name_is_correct(self):
        """Test that the span name is 'Evaluation output'."""
        expected_name = "Evaluation output"
        assert expected_name == "Evaluation output"

    def test_span_has_eval_output_span_type(self):
        """Test that span_type attribute is 'evalOutput'."""
        span_attributes = {"span.type": "evalOutput"}
        assert span_attributes["span.type"] == "evalOutput"

    def test_span_includes_value(self):
        """Test that value (score) is included in the span attributes."""
        score = 0.95
        span_attributes = {
            "span.type": "evalOutput",
            "value": score,
        }
        assert "value" in span_attributes
        assert span_attributes["value"] == 0.95

    def test_span_includes_evaluator_id(self):
        """Test that evaluatorId is included in the span attributes."""
        evaluator_id = "evaluator-123"
        span_attributes = {
            "span.type": "evalOutput",
            "evaluatorId": evaluator_id,
        }
        assert "evaluatorId" in span_attributes
        assert span_attributes["evaluatorId"] == evaluator_id

    def test_span_includes_justification(self):
        """Test that justification is included in the span attributes."""
        justification = "The output matches expected behavior."
        span_attributes = {
            "span.type": "evalOutput",
            "justification": justification,
        }
        assert "justification" in span_attributes
        assert span_attributes["justification"] == justification

    def test_span_has_all_required_attributes(self):
        """Test that all required attributes are present in the span."""
        evaluator_id = "eval-id-123"
        score = 100
        justification = "Perfect match"

        span_attributes = {
            "span.type": "evalOutput",
            "value": score,
            "evaluatorId": evaluator_id,
            "justification": justification,
        }

        # Verify all required attributes
        required_attrs = ["span.type", "value", "evaluatorId", "justification"]
        for attr in required_attrs:
            assert attr in span_attributes, f"Missing required attribute: {attr}"

    def test_span_has_openinference_kind(self):
        """Test that openinference.span.kind is set to CHAIN."""
        span_attributes = {
            "openinference.span.kind": "CHAIN",
            "span.type": "evalOutput",
        }
        assert span_attributes["openinference.span.kind"] == "CHAIN"

    def test_span_has_output_attribute_with_type_value_justification(self):
        """Test that span has output attribute with type, value, and justification."""
        import json

        # Simulate the output attribute set by set_evaluation_output_span_output
        output_data = {
            "type": 1,
            "value": 0.92,
            "justification": "The outputs are semantically equivalent",
        }
        output_json = json.dumps(output_data)

        span_attributes = {
            "span.type": "evalOutput",
            "output": output_json,
        }

        assert "output" in span_attributes
        parsed_output = json.loads(span_attributes["output"])
        assert parsed_output["type"] == 1
        assert parsed_output["value"] == 0.92
        assert (
            parsed_output["justification"] == "The outputs are semantically equivalent"
        )

    def test_span_output_type_is_always_one(self):
        """Test that output type field is always 1."""
        import json

        output_data = {"type": 1, "value": 0.5}
        output_json = json.dumps(output_data)

        span_attributes = {
            "span.type": "evalOutput",
            "output": output_json,
        }

        parsed_output = json.loads(span_attributes["output"])
        assert parsed_output["type"] == 1

    def test_span_output_without_justification(self):
        """Test that output can be set without justification field."""
        import json

        # When justification is None, it should be excluded from output
        output_data = {"type": 1, "value": 0.75}
        output_json = json.dumps(output_data)

        span_attributes = {
            "span.type": "evalOutput",
            "output": output_json,
        }

        parsed_output = json.loads(span_attributes["output"])
        assert parsed_output["type"] == 1
        assert parsed_output["value"] == 0.75
        assert "justification" not in parsed_output


class TestEvaluationOutputSpanHierarchy:
    """Tests verifying the Evaluation output span hierarchy."""

    def test_eval_output_is_child_of_evaluator(self):
        """Test that Evaluation output spans should be children of Evaluator spans."""
        parent_span_type = "evaluator"
        child_span_type = "evalOutput"

        # The parent-child relationship is enforced by span context nesting
        assert parent_span_type == "evaluator"
        assert child_span_type == "evalOutput"

    def test_eval_output_span_attributes_construction(self):
        """Test the construction of Evaluation output span attributes."""
        evaluator_id = "eval-123"
        score = 0.85
        justification = "Good accuracy"

        span_attributes = {
            "openinference.span.kind": "CHAIN",
            "span.type": "evalOutput",
            "value": score,
            "evaluatorId": evaluator_id,
            "justification": justification,
        }

        assert span_attributes["openinference.span.kind"] == "CHAIN"
        assert span_attributes["span.type"] == "evalOutput"
        assert span_attributes["value"] == 0.85
        assert span_attributes["evaluatorId"] == "eval-123"
        assert span_attributes["justification"] == "Good accuracy"


class TestEvaluationOutputSpanCreation:
    """Tests for Evaluation output span creation in progress reporter."""

    def test_eval_output_span_name_is_evaluation_output(self):
        """Test that the span name is exactly 'Evaluation output'."""
        span_name = "Evaluation output"
        assert span_name == "Evaluation output"

    def test_eval_output_span_type_is_camel_case(self):
        """Test that span.type uses camelCase: evalOutput."""
        span_type = "evalOutput"
        assert span_type == "evalOutput"
        # First letter lowercase, second word capitalized
        assert span_type[0].islower()
        assert "Output" in span_type

    def test_eval_output_with_pydantic_details(self):
        """Test that justification is extracted from Pydantic model details."""
        # Simulate Pydantic model details with justification field
        details_dict = {
            "justification": "The semantic similarity is perfect.",
            "other_field": "some value",
        }

        # Extract justification like the code does
        justification = details_dict.get("justification", str(details_dict))
        assert justification == "The semantic similarity is perfect."

    def test_eval_output_with_string_details(self):
        """Test that string details are used as justification directly."""
        details = "Good accuracy on all test cases"

        # String details are used directly
        justification = str(details)
        assert justification == "Good accuracy on all test cases"

    def test_eval_output_without_justification_field(self):
        """Test fallback when details dict has no justification field."""
        import json

        details_dict: dict[str, float] = {
            "accuracy": 0.95,
            "precision": 0.92,
        }

        # Should fall back to JSON dump of entire details
        # Since there's no "justification" key, we get the default JSON string
        justification = json.dumps(details_dict)
        assert "accuracy" in justification
        assert "0.95" in justification



import base64
import json
from os import environ as env
from pathlib import Path
from typing import Optional

from .constants import (
    ENV_BASE_URL,
    ENV_UIPATH_ACCESS_TOKEN,
    ENV_UNATTENDED_USER_ACCESS_TOKEN,
)


def parse_access_token(access_token: str):
    token_parts = access_token.split(".")
    if len(token_parts) < 2:
        raise Exception("Invalid access token")
    payload = base64.urlsafe_b64decode(
        token_parts[1] + "=" * (-len(token_parts[1]) % 4)
    )
    return json.loads(payload)


def update_env_file(env_contents):
    env_path = Path.cwd() / ".env"
    if env_path.exists():
        with open(env_path, "r") as f:
            for line in f:
                if "=" in line:
                    key, value = line.strip().split("=", 1)
                    if key not in env_contents:
                        env_contents[key] = value
    lines = [f"{key}={value}\n" for key, value in env_contents.items()]
    with open(env_path, "w") as f:
        f.writelines(lines)


def resolve_config_from_env(
    base_url: Optional[str],
    secret: Optional[str],
):
    """Simple config resolution from environment variables."""
    base_url_value = base_url or env.get(ENV_BASE_URL)
    secret_value = (
        secret
        or env.get(ENV_UNATTENDED_USER_ACCESS_TOKEN)
        or env.get(ENV_UIPATH_ACCESS_TOKEN)
    )
    return base_url_value, secret_value



"""End-to-end tests for input overrides functionality."""

import json
import tempfile
from pathlib import Path

import pytest

from uipath._cli._evals._models._evaluation_set import (
    EvaluationSet,
    LegacyEvaluationSet,
)
from uipath._cli._evals._runtime import UiPathEvalContext
from uipath._cli._utils._eval_set import EvalHelpers


@pytest.mark.asyncio
async def test_input_overrides_e2e_direct_override():
    """E2E test: per-evaluation input overrides on calculator inputs."""
    # Create a calculator agent.json
    agent_config = {
        "id": "test-calc-agent",
        "version": "1.0.0",
        "name": "Test Calculator",
        "messages": [
            {"role": "system", "content": "You are a calculator."},
            {"role": "user", "content": "Calculate: {{a}} {{operator}} {{b}}"},
        ],
        "inputSchema": {
            "type": "object",
            "properties": {
                "a": {"type": "number"},
                "b": {"type": "number"},
                "operator": {"type": "string"},
            },
        },
        "outputSchema": {
            "type": "object",
            "properties": {"result": {"type": "number"}},
        },
        "settings": {"model": "gpt-4", "temperature": 0},
    }

    # Create evaluation set with inputs that will be overridden
    eval_set = {
        "fileName": "test-eval.json",
        "id": "test-eval-set-id",
        "name": "Test Eval Set",
        "batchSize": 10,
        "evaluatorRefs": [],
        "evaluations": [
            {
                "id": "test-eval-1",
                "name": "Test Override",
                "inputs": {"a": 5, "b": 3, "operator": "+"},
                "expectedOutput": {"result": 50},  # Expect 10 * 5 = 50
                "expectedAgentBehavior": "Should calculate with overridden values",
                "evalSetId": "test-eval-set-id",
                "createdAt": "2025-12-16T00:00:00.000Z",
                "updatedAt": "2025-12-16T00:00:00.000Z",
            }
        ],
        "modelSettings": [],
        "createdAt": "2025-12-16T00:00:00.000Z",
        "updatedAt": "2025-12-16T00:00:00.000Z",
    }

    # Create temp files
    with tempfile.TemporaryDirectory() as tmpdir:
        tmpdir_path = Path(tmpdir)

        # Write agent.json
        agent_file = tmpdir_path / "agent.json"
        with open(agent_file, "w") as f:
            json.dump(agent_config, f)

        # Write eval set
        eval_file = tmpdir_path / "eval-set.json"
        with open(eval_file, "w") as f:
            json.dump(eval_set, f)

        # Load evaluation set to verify override structure
        loaded_eval_set, _ = EvalHelpers.load_eval_set(str(eval_file))
        assert isinstance(loaded_eval_set, (LegacyEvaluationSet, EvaluationSet))
        assert len(loaded_eval_set.evaluations) == 1
        original_input = loaded_eval_set.evaluations[0].inputs
        assert original_input["a"] == 5
        assert original_input["b"] == 3
        assert original_input["operator"] == "+"

        # Per-evaluation overrides
        overrides = {
            "test-eval-1": {
                "a": 10,
                "operator": "*",
            }
        }

        # Apply overrides to the inputs with eval_id
        from uipath._cli._evals._eval_util import apply_input_overrides

        modified_inputs = apply_input_overrides(
            original_input, overrides, eval_id="test-eval-1"
        )

        # Verify overrides were applied
        assert modified_inputs["a"] == 10  # Overridden
        assert modified_inputs["operator"] == "*"  # Overridden
        assert modified_inputs["b"] == 3  # Unchanged


@pytest.mark.asyncio
async def test_input_overrides_e2e_direct_field():
    """E2E test: input overrides with direct field override."""
    # Create evaluation set
    eval_set = {
        "fileName": "test-eval.json",
        "id": "test-eval-set-id",
        "name": "Test Eval Set",
        "batchSize": 10,
        "evaluatorRefs": [],
        "evaluations": [
            {
                "id": "test-eval-1",
                "name": "Test Direct Field Override",
                "inputs": {
                    "file_id": "old-attachment-123",
                    "document_id": "old-doc-456",
                    "count": 5,
                },
                "expectedOutput": {"status": "success"},
                "expectedAgentBehavior": "Process with new values",
                "evalSetId": "test-eval-set-id",
                "createdAt": "2025-12-16T00:00:00.000Z",
                "updatedAt": "2025-12-16T00:00:00.000Z",
            }
        ],
        "modelSettings": [],
        "createdAt": "2025-12-16T00:00:00.000Z",
        "updatedAt": "2025-12-16T00:00:00.000Z",
    }

    with tempfile.TemporaryDirectory() as tmpdir:
        tmpdir_path = Path(tmpdir)

        # Write eval set
        eval_file = tmpdir_path / "eval-set.json"
        with open(eval_file, "w") as f:
            json.dump(eval_set, f)

        # Load and test
        loaded_eval_set, _ = EvalHelpers.load_eval_set(str(eval_file))
        original_input = loaded_eval_set.evaluations[0].inputs

        # Per-evaluation overrides
        overrides = {
            "test-eval-1": {
                "file_id": "new-attachment-789",
                "document_id": "new-doc-999",
                "count": 10,
            }
        }

        from uipath._cli._evals._eval_util import apply_input_overrides

        modified_inputs = apply_input_overrides(
            original_input, overrides, eval_id="test-eval-1"
        )

        # Verify direct field override worked
        assert modified_inputs["file_id"] == "new-attachment-789"
        assert modified_inputs["document_id"] == "new-doc-999"
        assert modified_inputs["count"] == 10


@pytest.mark.asyncio
async def test_input_overrides_e2e_nested_objects():
    """E2E test: input overrides with nested objects using deep merge."""
    eval_set = {
        "fileName": "test-eval.json",
        "id": "test-eval-set-id",
        "name": "Test Eval Set",
        "batchSize": 10,
        "evaluatorRefs": [],
        "evaluations": [
            {
                "id": "test-eval-1",
                "name": "Test Nested Override",
                "inputs": {
                    "filePath": {
                        "ID": "old-attachment-id",
                        "FullName": "document.pdf",
                        "MimeType": "application/pdf",
                    },
                    "config": {"threshold": 0.8, "model": "gpt-4"},
                },
                "expectedOutput": {"status": "processed"},
                "expectedAgentBehavior": "Process with overridden attachment",
                "evalSetId": "test-eval-set-id",
                "createdAt": "2025-12-16T00:00:00.000Z",
                "updatedAt": "2025-12-16T00:00:00.000Z",
            }
        ],
        "modelSettings": [],
        "createdAt": "2025-12-16T00:00:00.000Z",
        "updatedAt": "2025-12-16T00:00:00.000Z",
    }

    with tempfile.TemporaryDirectory() as tmpdir:
        tmpdir_path = Path(tmpdir)
        eval_file = tmpdir_path / "eval-set.json"
        with open(eval_file, "w") as f:
            json.dump(eval_set, f)

        loaded_eval_set, _ = EvalHelpers.load_eval_set(str(eval_file))
        original_input = loaded_eval_set.evaluations[0].inputs

        # Per-evaluation overrides with deep merge for nested objects
        overrides = {
            "test-eval-1": {
                "filePath": {"ID": "new-attachment-id-xyz"},
                "config": {"threshold": 0.95},
            }
        }

        from uipath._cli._evals._eval_util import apply_input_overrides

        modified_inputs = apply_input_overrides(
            original_input, overrides, eval_id="test-eval-1"
        )

        # Verify deep merge - overridden fields updated, others preserved
        assert (
            modified_inputs["filePath"]["ID"] == "new-attachment-id-xyz"
        )  # Overridden
        assert modified_inputs["filePath"]["FullName"] == "document.pdf"  # Preserved
        assert modified_inputs["filePath"]["MimeType"] == "application/pdf"  # Preserved
        assert modified_inputs["config"]["threshold"] == 0.95  # Overridden
        assert modified_inputs["config"]["model"] == "gpt-4"  # Preserved


@pytest.mark.asyncio
async def test_input_overrides_context_field():
    """Test that UiPathEvalContext properly stores input_overrides."""
    context = UiPathEvalContext()
    context.input_overrides = {"eval-1": {"a": 10, "operator": "*"}}

    assert context.input_overrides is not None
    assert "eval-1" in context.input_overrides
    assert context.input_overrides["eval-1"]["a"] == 10
    assert context.input_overrides["eval-1"]["operator"] == "*"


@pytest.mark.asyncio
async def test_input_overrides_per_evaluation():
    """E2E test: per-evaluation input overrides."""
    # Per-evaluation overrides
    overrides = {
        "eval-1": {"operator": "*", "a": 10},
        "eval-2": {"operator": "/", "b": 5},
    }

    from uipath._cli._evals._eval_util import apply_input_overrides

    # Test eval-1
    eval_1_inputs = {"a": 5, "b": 3, "operator": "+"}
    modified_inputs_1 = apply_input_overrides(
        eval_1_inputs, overrides, eval_id="eval-1"
    )
    assert modified_inputs_1["a"] == 10  # Overridden
    assert modified_inputs_1["operator"] == "*"  # Overridden
    assert modified_inputs_1["b"] == 3  # Preserved

    # Test eval-2
    eval_2_inputs = {"a": 20, "b": 10, "operator": "+"}
    modified_inputs_2 = apply_input_overrides(
        eval_2_inputs, overrides, eval_id="eval-2"
    )
    assert modified_inputs_2["a"] == 20  # Preserved
    assert modified_inputs_2["b"] == 5  # Overridden
    assert modified_inputs_2["operator"] == "/"  # Overridden

    # Test eval-3 (no overrides)
    eval_3_inputs = {"a": 7, "b": 4, "operator": "-"}
    modified_inputs_3 = apply_input_overrides(
        eval_3_inputs, overrides, eval_id="eval-3"
    )
    assert modified_inputs_3["a"] == 7  # Unchanged
    assert modified_inputs_3["b"] == 4  # Unchanged
    assert modified_inputs_3["operator"] == "-"  # Unchanged



"""UiPath Connections Models.

This module contains models related to UiPath Connections service.
"""

from ._connections_service import ConnectionsService
from .connections import (
    ActivityMetadata,
    ActivityParameterLocationInfo,
    Connection,
    ConnectionMetadata,
    ConnectionToken,
    ConnectionTokenType,
    EventArguments,
)

__all__ = [
    "ConnectionsService",
    "ActivityMetadata",
    "ActivityParameterLocationInfo",
    "Connection",
    "ConnectionMetadata",
    "ConnectionToken",
    "ConnectionTokenType",
    "EventArguments",
]



import importlib
from pathlib import Path

import pytest

from uipath.platform import UiPathApiConfig, UiPathExecutionContext


@pytest.fixture
def base_url() -> str:
    return "https://test.uipath.com"


@pytest.fixture
def org() -> str:
    return "/org"


@pytest.fixture
def tenant() -> str:
    return "/tenant"


@pytest.fixture
def secret() -> str:
    return "secret"


@pytest.fixture
def config(base_url: str, org: str, tenant: str, secret: str) -> UiPathApiConfig:
    return UiPathApiConfig(base_url=f"{base_url}{org}{tenant}", secret=secret)


@pytest.fixture
def version(monkeypatch: pytest.MonkeyPatch) -> str:
    test_version = "1.0.0"
    monkeypatch.setattr(importlib.metadata, "version", lambda _: test_version)
    return test_version


@pytest.fixture
def execution_context(monkeypatch: pytest.MonkeyPatch) -> UiPathExecutionContext:
    monkeypatch.setenv("UIPATH_ROBOT_KEY", "test-robot-key")
    return UiPathExecutionContext()


@pytest.fixture
def tests_data_path() -> Path:
    return Path(__file__).resolve().parent / "tests_data"


@pytest.fixture
def jobs_service(config, execution_context):
    from uipath.platform.orchestrator import JobsService

    return JobsService(config, execution_context)



"""Tests for debug command tool simulation functionality."""

import json
from pathlib import Path
from typing import Any
from unittest.mock import Mock, patch

import pytest
from click.testing import CliRunner

from uipath._cli import cli
from uipath._cli._evals.mocks.mocks import (
    clear_execution_context,
    is_tool_simulated,
)
from uipath._cli._evals.mocks.types import (
    LLMMockingStrategy,
    MockingContext,
    MockingStrategyType,
    ToolSimulation,
)
from uipath._cli.cli_debug import load_simulation_config
from uipath._cli.middlewares import MiddlewareResult


@pytest.fixture
def valid_simulation_config() -> dict[str, Any]:
    """Create a valid simulation.json configuration."""
    return {
        "instructions": "Always give a negative outlook on stock prospects",
        "inputGenerationInstructions": "",
        "simulateInput": False,
        "enabled": True,
        "toolsToSimulate": [
            {"name": "Web Reader"},
            {"name": "Web Search"},
            {"name": "Web Summary"},
        ],
    }


@pytest.fixture
def disabled_simulation_config() -> dict[str, Any]:
    """Create a disabled simulation.json configuration."""
    return {
        "instructions": "Test instructions",
        "enabled": False,
        "toolsToSimulate": [{"name": "Test Tool"}],
    }


@pytest.fixture
def empty_tools_simulation_config() -> dict[str, Any]:
    """Create simulation config with no tools."""
    return {
        "instructions": "Test instructions",
        "enabled": True,
        "toolsToSimulate": [],
    }


class TestLoadSimulationConfig:
    """Tests for the load_simulation_config function."""

    def test_returns_none_when_file_does_not_exist(self, temp_dir: str):
        """Test that None is returned when simulation.json doesn't exist."""
        with patch("uipath._cli.cli_debug.Path.cwd", return_value=Path(temp_dir)):
            result = load_simulation_config()
            assert result is None

    def test_loads_valid_simulation_config(
        self, temp_dir: str, valid_simulation_config: dict[str, Any]
    ):
        """Test loading a valid simulation configuration."""
        simulation_path = Path(temp_dir) / "simulation.json"
        with open(simulation_path, "w", encoding="utf-8") as f:
            json.dump(valid_simulation_config, f)

        with patch("uipath._cli.cli_debug.Path.cwd", return_value=Path(temp_dir)):
            result = load_simulation_config()

            assert result is not None
            assert isinstance(result, MockingContext)
            assert result.name == "debug-simulation"
            assert result.strategy is not None
            assert isinstance(result.strategy, LLMMockingStrategy)
            assert result.strategy.prompt == valid_simulation_config["instructions"]
            assert len(result.strategy.tools_to_simulate) == 3
            assert result.strategy.tools_to_simulate[0].name == "Web Reader"

    def test_returns_none_when_disabled(
        self, temp_dir: str, disabled_simulation_config: dict[str, Any]
    ):
        """Test that None is returned when simulation is disabled."""
        simulation_path = Path(temp_dir) / "simulation.json"
        with open(simulation_path, "w", encoding="utf-8") as f:
            json.dump(disabled_simulation_config, f)

        with patch("uipath._cli.cli_debug.Path.cwd", return_value=Path(temp_dir)):
            result = load_simulation_config()
            assert result is None

    def test_returns_none_when_no_tools_to_simulate(
        self, temp_dir: str, empty_tools_simulation_config: dict[str, Any]
    ):
        """Test that None is returned when toolsToSimulate is empty."""
        simulation_path = Path(temp_dir) / "simulation.json"
        with open(simulation_path, "w", encoding="utf-8") as f:
            json.dump(empty_tools_simulation_config, f)

        with patch("uipath._cli.cli_debug.Path.cwd", return_value=Path(temp_dir)):
            result = load_simulation_config()
            assert result is None

    def test_handles_malformed_json(self, temp_dir: str):
        """Test that malformed JSON is handled gracefully."""
        simulation_path = Path(temp_dir) / "simulation.json"
        with open(simulation_path, "w", encoding="utf-8") as f:
            f.write("{ invalid json }")

        with patch("uipath._cli.cli_debug.Path.cwd", return_value=Path(temp_dir)):
            result = load_simulation_config()
            assert result is None

    def test_handles_missing_required_fields(self, temp_dir: str):
        """Test that missing required fields are handled gracefully."""
        simulation_path = Path(temp_dir) / "simulation.json"
        with open(simulation_path, "w", encoding="utf-8") as f:
            json.dump({"enabled": True}, f)

        with patch("uipath._cli.cli_debug.Path.cwd", return_value=Path(temp_dir)):
            result = load_simulation_config()
            # Should return None because toolsToSimulate is missing/empty
            assert result is None

    def test_creates_mocking_context_with_empty_inputs(
        self, temp_dir: str, valid_simulation_config: dict[str, Any]
    ):
        """Test that MockingContext is created with empty inputs."""
        simulation_path = Path(temp_dir) / "simulation.json"
        with open(simulation_path, "w", encoding="utf-8") as f:
            json.dump(valid_simulation_config, f)

        with patch("uipath._cli.cli_debug.Path.cwd", return_value=Path(temp_dir)):
            result = load_simulation_config()

            assert result is not None
            assert result.inputs == {}

    def test_uses_default_empty_instructions_when_missing(self, temp_dir: str):
        """Test that empty string is used when instructions field is missing."""
        config = {
            "enabled": True,
            "toolsToSimulate": [{"name": "Test Tool"}],
        }
        simulation_path = Path(temp_dir) / "simulation.json"
        with open(simulation_path, "w", encoding="utf-8") as f:
            json.dump(config, f)

        with patch("uipath._cli.cli_debug.Path.cwd", return_value=Path(temp_dir)):
            result = load_simulation_config()

            assert result is not None
            assert isinstance(result.strategy, LLMMockingStrategy)
            assert result.strategy.prompt == ""


class TestDebugCommandSimulationIntegration:
    """Integration tests for debug command with simulation."""

    @pytest.fixture
    def mock_runtime(self):
        """Create a mock runtime for testing."""
        runtime = Mock()
        runtime.execute = Mock(return_value=Mock(status="SUCCESSFUL", output={}))
        runtime.dispose = Mock()
        return runtime

    @pytest.fixture
    def mock_factory(self, mock_runtime):
        """Create a mock factory that returns mock runtime."""
        factory = Mock()
        factory.new_runtime = Mock(return_value=mock_runtime)
        factory.dispose = Mock()
        return factory

    def test_debug_without_simulation_file(self, runner: CliRunner, temp_dir: str):
        """Test debug command when simulation.json doesn't exist."""
        with runner.isolated_filesystem(temp_dir=temp_dir):
            script_file = "entrypoint.py"
            with open(script_file, "w") as f:
                f.write("def main(input): return {'result': 'success'}")

            # Create uipath.json
            with open("uipath.json", "w") as f:
                json.dump({"functions": {"main": f"{script_file}:main"}}, f)

            with patch("uipath._cli.cli_debug.Middlewares.next") as mock_middleware:
                mock_middleware.return_value = MiddlewareResult(
                    should_continue=False,
                    info_message="Execution succeeded",
                    error_message=None,
                    should_include_stacktrace=False,
                )
                result = runner.invoke(cli, ["debug", "main", "{}"])
                assert result.exit_code == 0

    def test_debug_with_simulation_file_sets_context(
        self, runner: CliRunner, temp_dir: str, valid_simulation_config: dict[str, Any]
    ):
        """Test that debug command sets execution context when simulation.json exists."""
        with runner.isolated_filesystem(temp_dir=temp_dir):
            script_file = "entrypoint.py"
            with open(script_file, "w") as f:
                f.write(
                    """
from uipath.eval.mocks import mockable

@mockable(name="Web Reader")
def web_reader(url: str) -> str:
    return "Original content"

def main(input):
    return {'result': web_reader('https://test.com')}
"""
                )

            # Create uipath.json
            with open("uipath.json", "w") as f:
                json.dump({"functions": {"main": f"{script_file}:main"}}, f)

            # Create a MockingContext to return from load_simulation_config
            mocking_strategy = LLMMockingStrategy(
                type=MockingStrategyType.LLM,
                prompt=valid_simulation_config["instructions"],
                tools_to_simulate=[
                    ToolSimulation(name=tool["name"])
                    for tool in valid_simulation_config["toolsToSimulate"]
                ],
            )
            mock_mocking_context = MockingContext(
                strategy=mocking_strategy,
                name="debug-simulation",
                inputs={},
            )

            # Track if set_execution_context was called
            with patch(
                "uipath._cli.cli_debug.set_execution_context"
            ) as mock_set_context:
                with patch(
                    "uipath._cli.cli_debug.clear_execution_context"
                ) as mock_clear_context:
                    # Mock load_simulation_config to return the MockingContext
                    with patch(
                        "uipath._cli.cli_debug.load_simulation_config"
                    ) as mock_load_config:
                        mock_load_config.return_value = mock_mocking_context

                        with patch(
                            "uipath._cli.cli_debug.Middlewares.next"
                        ) as mock_middleware:
                            # Set should_continue=True so the debug execution logic runs
                            mock_middleware.return_value = MiddlewareResult(
                                should_continue=True,
                                info_message=None,
                                error_message=None,
                                should_include_stacktrace=False,
                            )

                            # Mock the runtime factory and execution
                            with patch(
                                "uipath._cli.cli_debug.UiPathRuntimeFactoryRegistry.get"
                            ) as mock_factory_get:
                                mock_runtime = Mock()
                                mock_runtime.execute = Mock(
                                    return_value=Mock(status="SUCCESSFUL", output={})
                                )
                                mock_runtime.dispose = Mock()

                                mock_factory = Mock()
                                mock_factory.new_runtime = Mock(
                                    return_value=mock_runtime
                                )
                                mock_factory.dispose = Mock()
                                mock_factory_get.return_value = mock_factory

                                # Mock debug bridge to avoid SignalR/console issues
                                with patch("uipath._cli.cli_debug.get_debug_bridge"):
                                    with patch(
                                        "uipath._cli.cli_debug.UiPathDebugRuntime"
                                    ) as mock_debug_runtime_class:
                                        mock_debug_runtime = Mock()
                                        mock_debug_runtime.execute = Mock(
                                            return_value=Mock(
                                                status="SUCCESSFUL", output={}
                                            )
                                        )
                                        mock_debug_runtime.dispose = Mock()
                                        mock_debug_runtime_class.return_value = (
                                            mock_debug_runtime
                                        )

                                        runner.invoke(cli, ["debug", "main", "{}"])

                                        # Verify set_execution_context was called
                                        assert mock_set_context.called
                                        # Verify the MockingContext passed has the right structure
                                        call_args = mock_set_context.call_args
                                        mocking_ctx = call_args[0][0]
                                        assert isinstance(mocking_ctx, MockingContext)
                                        assert mocking_ctx.strategy is not None
                                        assert isinstance(
                                            mocking_ctx.strategy, LLMMockingStrategy
                                        )
                                        assert (
                                            len(mocking_ctx.strategy.tools_to_simulate)
                                            == 3
                                        )

                                        # Verify clear_execution_context was called in finally block
                                        assert mock_clear_context.called

    def test_debug_clears_context_on_error(
        self, runner: CliRunner, temp_dir: str, valid_simulation_config: dict[str, Any]
    ):
        """Test that execution context is cleared even when an error occurs."""
        with runner.isolated_filesystem(temp_dir=temp_dir):
            # Create simulation.json
            with open("simulation.json", "w") as f:
                json.dump(valid_simulation_config, f)

            script_file = "entrypoint.py"
            with open(script_file, "w") as f:
                f.write("def main(input): raise Exception('Test error')")

            # Create uipath.json
            with open("uipath.json", "w") as f:
                json.dump({"functions": {"main": f"{script_file}:main"}}, f)

            # Create a MockingContext to return from load_simulation_config
            mocking_strategy = LLMMockingStrategy(
                type=MockingStrategyType.LLM,
                prompt="Test instructions",
                tools_to_simulate=[ToolSimulation(name="Test Tool")],
            )
            mock_mocking_context = MockingContext(
                strategy=mocking_strategy,
                name="test-simulation",
                inputs={},
            )

            with patch(
                "uipath._cli.cli_debug.clear_execution_context"
            ) as mock_clear_context:
                with patch(
                    "uipath._cli.cli_debug.load_simulation_config"
                ) as mock_load_config:
                    # Mock load_simulation_config to return a MockingContext
                    mock_load_config.return_value = mock_mocking_context

                    with patch(
                        "uipath._cli.cli_debug.Middlewares.next"
                    ) as mock_middleware:
                        mock_middleware.return_value = MiddlewareResult(
                            should_continue=True,
                            info_message=None,
                            error_message=None,
                            should_include_stacktrace=False,
                        )

                        # Mock the runtime factory and execution to simulate an error
                        with patch(
                            "uipath._cli.cli_debug.UiPathRuntimeFactoryRegistry.get"
                        ) as mock_factory_get:
                            mock_runtime = Mock()
                            # Make execute raise an exception to simulate an error
                            mock_runtime.execute = Mock(
                                side_effect=Exception("Test error during execution")
                            )
                            mock_runtime.dispose = Mock()

                            mock_factory = Mock()
                            mock_factory.new_runtime = Mock(return_value=mock_runtime)
                            mock_factory.dispose = Mock()
                            mock_factory_get.return_value = mock_factory

                            # Mock debug bridge to avoid SignalR/console issues
                            with patch("uipath._cli.cli_debug.get_debug_bridge"):
                                with patch(
                                    "uipath._cli.cli_debug.UiPathDebugRuntime"
                                ) as mock_debug_runtime_class:
                                    mock_debug_runtime = Mock()
                                    # Make the debug runtime raise an exception
                                    mock_debug_runtime.execute = Mock(
                                        side_effect=Exception("Test error")
                                    )
                                    mock_debug_runtime.dispose = Mock()
                                    mock_debug_runtime_class.return_value = (
                                        mock_debug_runtime
                                    )

                                    # This will raise an exception during execution
                                    runner.invoke(cli, ["debug", "main", "{}"])

                                    # Verify clear_execution_context was still called
                                    assert mock_clear_context.called

    def test_simulation_config_enables_tool_mocking(
        self, temp_dir: str, valid_simulation_config: dict[str, Any]
    ):
        """Test that tools marked in simulation config are detected as simulated."""
        # Clear any existing context
        clear_execution_context()

        simulation_path = Path(temp_dir) / "simulation.json"
        with open(simulation_path, "w", encoding="utf-8") as f:
            json.dump(valid_simulation_config, f)

        with patch("uipath._cli.cli_debug.Path.cwd", return_value=Path(temp_dir)):
            mocking_ctx = load_simulation_config()
            assert mocking_ctx is not None

            # Manually set context (simulating what debug command does)
            from uipath._cli._evals._span_collection import ExecutionSpanCollector

            span_collector = ExecutionSpanCollector()
            from uipath._cli._evals.mocks.mocks import set_execution_context

            set_execution_context(mocking_ctx, span_collector, "test-execution-id")

            # Verify tools are detected as simulated
            assert is_tool_simulated("Web Reader") is True
            assert is_tool_simulated("Web Search") is True
            assert is_tool_simulated("Web Summary") is True
            assert is_tool_simulated("Non Simulated Tool") is False

            # Clean up
            clear_execution_context()

    def test_disabled_simulation_does_not_set_context(
        self,
        runner: CliRunner,
        temp_dir: str,
        disabled_simulation_config: dict[str, Any],
    ):
        """Test that disabled simulation doesn't set execution context."""
        with runner.isolated_filesystem(temp_dir=temp_dir):
            # Create disabled simulation.json
            with open("simulation.json", "w") as f:
                json.dump(disabled_simulation_config, f)

            script_file = "entrypoint.py"
            with open(script_file, "w") as f:
                f.write("def main(input): return {'result': 'success'}")

            # Create uipath.json
            with open("uipath.json", "w") as f:
                json.dump({"functions": {"main": f"{script_file}:main"}}, f)

            with patch(
                "uipath._cli.cli_debug.set_execution_context"
            ) as mock_set_context:
                with patch("uipath._cli.cli_debug.Middlewares.next") as mock_middleware:
                    mock_middleware.return_value = MiddlewareResult(
                        should_continue=False,
                        info_message="Execution succeeded",
                        error_message=None,
                        should_include_stacktrace=False,
                    )

                    runner.invoke(cli, ["debug", "main", "{}"])

                    # Verify set_execution_context was NOT called
                    assert not mock_set_context.called


class TestSimulationConfigFields:
    """Test various field combinations in simulation.json."""

    def test_enabled_defaults_to_true_when_missing(self, temp_dir: str):
        """Test that 'enabled' defaults to true when not specified."""
        config = {
            "instructions": "Test",
            "toolsToSimulate": [{"name": "Test Tool"}],
        }
        simulation_path = Path(temp_dir) / "simulation.json"
        with open(simulation_path, "w", encoding="utf-8") as f:
            json.dump(config, f)

        with patch("uipath._cli.cli_debug.Path.cwd", return_value=Path(temp_dir)):
            result = load_simulation_config()
            # Should load successfully since enabled defaults to true
            assert result is not None

    def test_handles_tool_name_normalization(self, temp_dir: str):
        """Test that tool names with underscores work correctly."""
        config = {
            "enabled": True,
            "instructions": "Test",
            "toolsToSimulate": [
                {"name": "Web_Reader"},
                {"name": "Web Search"},
            ],
        }
        simulation_path = Path(temp_dir) / "simulation.json"
        with open(simulation_path, "w", encoding="utf-8") as f:
            json.dump(config, f)

        with patch("uipath._cli.cli_debug.Path.cwd", return_value=Path(temp_dir)):
            mocking_ctx = load_simulation_config()
            assert mocking_ctx is not None

            # Set context to test name normalization
            from uipath._cli._evals._span_collection import ExecutionSpanCollector
            from uipath._cli._evals.mocks.mocks import set_execution_context

            span_collector = ExecutionSpanCollector()
            set_execution_context(mocking_ctx, span_collector, "test-id")

            # Both underscore and space versions should be detected
            assert is_tool_simulated("Web_Reader") is True
            assert is_tool_simulated("Web Reader") is True
            assert is_tool_simulated("Web Search") is True

            clear_execution_context()



"""UiPath evaluator implementations for agent performance evaluation."""

from typing import Any

# Current coded evaluators
from .base_evaluator import BaseEvaluationCriteria, BaseEvaluator, BaseEvaluatorConfig
from .base_legacy_evaluator import BaseLegacyEvaluator

# Legacy evaluators
from .contains_evaluator import ContainsEvaluator
from .exact_match_evaluator import ExactMatchEvaluator
from .json_similarity_evaluator import JsonSimilarityEvaluator
from .legacy_context_precision_evaluator import LegacyContextPrecisionEvaluator
from .legacy_exact_match_evaluator import LegacyExactMatchEvaluator
from .legacy_faithfulness_evaluator import LegacyFaithfulnessEvaluator
from .legacy_json_similarity_evaluator import LegacyJsonSimilarityEvaluator
from .legacy_llm_as_judge_evaluator import LegacyLlmAsAJudgeEvaluator
from .legacy_trajectory_evaluator import LegacyTrajectoryEvaluator
from .llm_judge_output_evaluator import (
    BaseLLMOutputEvaluator,
    LLMJudgeOutputEvaluator,
    LLMJudgeStrictJSONSimilarityOutputEvaluator,
)
from .llm_judge_trajectory_evaluator import (
    BaseLLMTrajectoryEvaluator,
    LLMJudgeTrajectoryEvaluator,
    LLMJudgeTrajectorySimulationEvaluator,
)
from .tool_call_args_evaluator import ToolCallArgsEvaluator
from .tool_call_count_evaluator import ToolCallCountEvaluator
from .tool_call_order_evaluator import ToolCallOrderEvaluator
from .tool_call_output_evaluator import ToolCallOutputEvaluator

EVALUATORS: list[type[BaseEvaluator[Any, Any, Any]]] = [
    ExactMatchEvaluator,
    ContainsEvaluator,
    JsonSimilarityEvaluator,
    LLMJudgeOutputEvaluator,
    LLMJudgeStrictJSONSimilarityOutputEvaluator,
    LLMJudgeTrajectoryEvaluator,
    LLMJudgeTrajectorySimulationEvaluator,
    ToolCallOrderEvaluator,
    ToolCallArgsEvaluator,
    ToolCallCountEvaluator,
    ToolCallOutputEvaluator,
]
__all__ = [
    # Legacy evaluators
    "BaseLegacyEvaluator",
    "LegacyContextPrecisionEvaluator",
    "LegacyExactMatchEvaluator",
    "LegacyFaithfulnessEvaluator",
    "LegacyLlmAsAJudgeEvaluator",
    "LegacyTrajectoryEvaluator",
    "LegacyJsonSimilarityEvaluator",
    # Current coded evaluators
    "BaseEvaluator",
    "ContainsEvaluator",
    "ExactMatchEvaluator",
    "JsonSimilarityEvaluator",
    "BaseLLMOutputEvaluator",
    "LLMJudgeOutputEvaluator",
    "LLMJudgeStrictJSONSimilarityOutputEvaluator",
    "BaseLLMTrajectoryEvaluator",
    "LLMJudgeTrajectoryEvaluator",
    "LLMJudgeTrajectorySimulationEvaluator",
    "ToolCallOrderEvaluator",
    "ToolCallArgsEvaluator",
    "ToolCallCountEvaluator",
    "ToolCallOutputEvaluator",
    "BaseEvaluationCriteria",
    "BaseEvaluatorConfig",
]



"""Type conversion utilities for runtime input/output handling."""

import inspect
from dataclasses import asdict, is_dataclass
from typing import Any, Type, TypeVar, cast, get_type_hints

from pydantic import BaseModel

T = TypeVar("T")


def is_pydantic_model(cls: Type[Any]) -> bool:
    """Check if a class is a Pydantic model."""
    try:
        return inspect.isclass(cls) and issubclass(cls, BaseModel)
    except TypeError:
        return False


def convert_to_class(data: dict[str, Any], cls: Type[T]) -> T:
    """Convert a dictionary to a class instance (Pydantic, dataclass, or regular)."""
    # Pydantic models
    if is_pydantic_model(cls):
        pydantic_cls = cast(Type[BaseModel], cls)
        return cast(T, pydantic_cls.model_validate(data))

    # Dataclasses
    if is_dataclass(cls):
        return _convert_to_dataclass(data, cls)

    # Regular classes
    return _convert_to_regular_class(data, cls)


def _convert_to_dataclass(data: dict[str, Any], cls: Type[T]) -> T:
    """Convert dictionary to dataclass instance."""
    field_types = get_type_hints(cls)
    converted_data = {}

    for field_name, field_type in field_types.items():
        if field_name not in data:
            continue

        value = data[field_name]
        # Recursively convert nested structures
        if isinstance(value, dict) and _is_nested_type(field_type):
            value = convert_to_class(value, cast(Type[Any], field_type))

        converted_data[field_name] = value

    return cls(**converted_data)


def _convert_to_regular_class(data: dict[str, Any], cls: Type[T]) -> T:
    """Convert dictionary to regular class instance."""
    sig = inspect.signature(cls.__init__)
    init_args = {}

    for param_name, param in sig.parameters.items():
        if param_name == "self":
            continue

        if param_name in data:
            value = data[param_name]
            param_type = (
                param.annotation if param.annotation != inspect.Parameter.empty else Any
            )

            # Recursively convert nested structures
            if isinstance(value, dict) and _is_nested_type(param_type):
                value = convert_to_class(value, cast(Type[Any], param_type))

            init_args[param_name] = value
        elif param.default != inspect.Parameter.empty:
            init_args[param_name] = param.default

    return cls(**init_args)


def convert_from_class(obj: Any) -> dict[str, Any]:
    """Convert a class instance to a dictionary."""
    if obj is None:
        return {}

    # Handle primitives and built-in types first
    if isinstance(obj, (str, int, float, bool, list, tuple)):
        return {"result": obj}

    # Handle dict - return as-is
    if isinstance(obj, dict):
        return obj

    # Pydantic models
    if isinstance(obj, BaseModel):
        return obj.model_dump()

    # Dataclasses
    if is_dataclass(obj) and not isinstance(obj, type):
        return asdict(obj)

    # Regular classes with __dict__
    if hasattr(obj, "__dict__") and not isinstance(obj, type):
        return _convert_from_regular_class(obj)

    # Fallback
    return {}


def _convert_from_regular_class(obj: Any) -> dict[str, Any]:
    """Convert regular class instance to dictionary."""
    result = {}
    for key, value in obj.__dict__.items():
        if not key.startswith("_"):
            # Recursively convert nested objects
            if (
                isinstance(value, BaseModel)
                or is_dataclass(value)
                or (
                    hasattr(value, "__dict__")
                    and not isinstance(value, (dict, list, str, int, float, bool))
                )
            ):
                result[key] = convert_from_class(value)
            else:
                result[key] = value
    return result


def _is_nested_type(type_hint: Type[Any]) -> bool:
    """Check if a type is a nested structure that needs conversion."""
    return (
        is_dataclass(type_hint)
        or is_pydantic_model(type_hint)
        or (inspect.isclass(type_hint) and hasattr(type_hint, "__annotations__"))
    )



"""Tests for evaluator evaluate() methods.

This module tests the actual evaluation functionality of all evaluators:
- ExactMatchEvaluator.evaluate()
- JsonSimilarityEvaluator.evaluate()
- LlmAsAJudgeEvaluator.evaluate()
- ToolCallOrderEvaluator.evaluate()
- ToolCallCountEvaluator.evaluate()
- LlmJudgeTrajectoryEvaluator.evaluate()
"""

import math
import uuid
from typing import Any

import pytest
from opentelemetry.sdk.trace import ReadableSpan
from pydantic import ValidationError
from pytest_mock.plugin import MockerFixture

from uipath.eval.evaluators.contains_evaluator import (
    ContainsEvaluationCriteria,
    ContainsEvaluator,
)
from uipath.eval.evaluators.exact_match_evaluator import ExactMatchEvaluator
from uipath.eval.evaluators.json_similarity_evaluator import (
    JsonSimilarityEvaluator,
)
from uipath.eval.evaluators.llm_judge_output_evaluator import (
    LLMJudgeOutputEvaluator,
)
from uipath.eval.evaluators.llm_judge_trajectory_evaluator import (
    LLMJudgeTrajectoryEvaluator,
    TrajectoryEvaluationCriteria,
)
from uipath.eval.evaluators.output_evaluator import OutputEvaluationCriteria
from uipath.eval.evaluators.tool_call_args_evaluator import (
    ToolCallArgsEvaluationCriteria,
    ToolCallArgsEvaluator,
)
from uipath.eval.evaluators.tool_call_count_evaluator import (
    ToolCallCountEvaluationCriteria,
    ToolCallCountEvaluator,
)
from uipath.eval.evaluators.tool_call_order_evaluator import (
    ToolCallOrderEvaluationCriteria,
    ToolCallOrderEvaluator,
)
from uipath.eval.evaluators.tool_call_output_evaluator import (
    ToolCallOutputEvaluationCriteria,
    ToolCallOutputEvaluator,
    ToolCallOutputEvaluatorJustification,
)
from uipath.eval.models import NumericEvaluationResult
from uipath.eval.models.models import (
    AgentExecution,
    ToolCall,
    ToolOutput,
    UiPathEvaluationError,
)


@pytest.fixture
def sample_agent_execution() -> AgentExecution:
    """Create a sample AgentExecution for testing."""
    return AgentExecution(
        agent_input={"input": "Test input"},
        agent_output={"output": "Test output"},
        agent_trace=[],  # Empty trace for basic tests
    )


@pytest.fixture
def sample_agent_execution_with_trace() -> AgentExecution:
    """Create a sample AgentExecution with tool call trace."""
    # Mock spans that represent tool calls - simplified for testing
    mock_spans = [
        ReadableSpan(
            name="tool1",
            start_time=0,
            end_time=1,
            attributes={
                "tool.name": "tool1",
                "input.value": "{'arg1': 'value1'}",
                "output.value": '{"content": "output1"}',
            },
        ),
        ReadableSpan(
            name="tool2",
            start_time=1,
            end_time=2,
            attributes={
                "tool.name": "tool2",
                "input.value": "{'arg2': 'value2'}",
                "output.value": '{"content": "output2"}',
            },
        ),
        ReadableSpan(
            name="tool1",
            start_time=2,
            end_time=3,
            attributes={
                "tool.name": "tool1",
                "input.value": "{'arg1': 'value1'}",
                "output.value": '{"content": "output1"}',
            },
        ),
        ReadableSpan(
            name="tool2",
            start_time=3,
            end_time=4,
            attributes={
                "tool.name": "tool2",
                "input.value": "{'arg2': 'value2'}",
                "output.value": '{"content": "output2"}',
            },
        ),
    ]

    return AgentExecution(
        agent_input={"input": "Test input with tools"},
        agent_output={
            "output": "Test output with tools",
        },
        agent_trace=mock_spans,
    )


class TestExactMatchEvaluator:
    """Test ExactMatchEvaluator.evaluate() method."""

    @pytest.mark.asyncio
    async def test_exact_match_string_success(
        self, sample_agent_execution: AgentExecution
    ) -> None:
        """Test exact match with matching strings."""
        config = {
            "name": "ExactMatchTest",
            "case_sensitive": True,
            "default_evaluation_criteria": {"expected_output": "test"},
        }
        evaluator = ExactMatchEvaluator.model_validate(
            {"evaluatorConfig": config, "id": str(uuid.uuid4())}
        )
        criteria = OutputEvaluationCriteria(expected_output={"output": "Test output"})  # pyright: ignore[reportCallIssue]

        result = await evaluator.evaluate(sample_agent_execution, criteria)

        assert isinstance(result, NumericEvaluationResult)
        assert result.score == 1.0

    @pytest.mark.asyncio
    async def test_exact_match_string_failure(
        self, sample_agent_execution: AgentExecution
    ) -> None:
        """Test exact match with non-matching strings."""
        config = {
            "name": "ExactMatchTest",
            "case_sensitive": True,
            "default_evaluation_criteria": {"expected_output": "test"},
        }
        evaluator = ExactMatchEvaluator.model_validate(
            {"evaluatorConfig": config, "id": str(uuid.uuid4())}
        )
        criteria = OutputEvaluationCriteria(
            expected_output={"output": "Different output"}  # pyright: ignore[reportCallIssue]
        )

        result = await evaluator.evaluate(sample_agent_execution, criteria)

        assert isinstance(result, NumericEvaluationResult)
        assert result.score == 0.0

    @pytest.mark.asyncio
    async def test_exact_match_negated(
        self, sample_agent_execution: AgentExecution
    ) -> None:
        """Test exact match with negated criteria."""
        config = {
            "name": "ExactMatchTest",
            "case_sensitive": True,
            "negated": True,
        }
        evaluator = ExactMatchEvaluator.model_validate(
            {"evaluatorConfig": config, "id": str(uuid.uuid4())}
        )
        criteria = OutputEvaluationCriteria(
            expected_output={"output": "Test output"},  # pyright: ignore[reportCallIssue]
        )

        result = await evaluator.evaluate(sample_agent_execution, criteria)

        assert isinstance(result, NumericEvaluationResult)
        assert result.score == 0.0

    @pytest.mark.asyncio
    async def test_exact_match_validate_and_evaluate_criteria(
        self, sample_agent_execution: AgentExecution
    ) -> None:
        """Test exact match using validate_and_evaluate_criteria."""
        config = {
            "name": "ExactMatchTest",
            "case_sensitive": True,
        }
        evaluator = ExactMatchEvaluator.model_validate(
            {"evaluatorConfig": config, "id": str(uuid.uuid4())}
        )
        raw_criteria = {"expected_output": {"output": "Test output"}}

        result = await evaluator.validate_and_evaluate_criteria(
            sample_agent_execution, raw_criteria
        )

        assert isinstance(result, NumericEvaluationResult)
        assert result.score == 1.0


class TestContainsEvaluator:
    """Test ContainsEvaluator.evaluate() method."""

    @pytest.mark.asyncio
    async def test_contains_evaluator(
        self, sample_agent_execution: AgentExecution
    ) -> None:
        """Test contains evaluator."""
        config = {
            "name": "ContainsTest",
            "target_output_key": "output",
            "default_evaluation_criteria": {"search_text": "Test output"},
        }
        evaluator = ContainsEvaluator.model_validate(
            {"evaluatorConfig": config, "id": str(uuid.uuid4())}
        )
        criteria = ContainsEvaluationCriteria(search_text="Test output")
        result = await evaluator.evaluate(sample_agent_execution, criteria)

        assert isinstance(result, NumericEvaluationResult)
        assert result.score == 1.0

    @pytest.mark.asyncio
    async def test_contains_evaluator_negated(
        self, sample_agent_execution: AgentExecution
    ) -> None:
        """Test contains evaluator with negated criteria."""
        config = {
            "name": "ContainsTest",
            "negated": True,
            "target_output_key": "output",
            "default_evaluation_criteria": {"search_text": "Test output"},
        }
        evaluator = ContainsEvaluator.model_validate(
            {"evaluatorConfig": config, "id": str(uuid.uuid4())}
        )
        criteria = ContainsEvaluationCriteria(search_text="Test output")
        result = await evaluator.evaluate(sample_agent_execution, criteria)

        assert isinstance(result, NumericEvaluationResult)
        assert result.score == 0.0

    @pytest.mark.asyncio
    async def test_contains_evaluator_validate_and_evaluate_criteria(
        self, sample_agent_execution: AgentExecution
    ) -> None:
        """Test contains evaluator with validate_and_evaluate_criteria."""
        config = {
            "name": "ContainsTest",
            "target_output_key": "*",
            "default_evaluation_criteria": {"search_text": "Test output"},
        }
        evaluator = ContainsEvaluator.model_validate(
            {"evaluatorConfig": config, "id": str(uuid.uuid4())}
        )
        criteria = ContainsEvaluationCriteria(search_text="Test output")
        result = await evaluator.validate_and_evaluate_criteria(
            sample_agent_execution, criteria
        )
        assert isinstance(result, NumericEvaluationResult)
        assert result.score == 1.0


class TestJsonSimilarityEvaluator:
    """Test JsonSimilarityEvaluator.evaluate() method."""

    @pytest.mark.asyncio
    async def test_json_similarity_identical(self) -> None:
        """Test JSON similarity with identical structures."""
        execution = AgentExecution(
            agent_input={"input": "Test"},
            agent_output={"name": "John", "age": 30, "city": "NYC"},
            agent_trace=[],
        )
        config = {
            "name": "JsonSimilarityTest",
        }
        evaluator = JsonSimilarityEvaluator.model_validate(
            {"evaluatorConfig": config, "id": str(uuid.uuid4())}
        )
        criteria = OutputEvaluationCriteria(
            expected_output={"name": "John", "age": 30, "city": "NYC"}  # pyright: ignore[reportCallIssue]
        )

        result = await evaluator.evaluate(execution, criteria)

        assert isinstance(result, NumericEvaluationResult)
        assert result.score == 1.0

    @pytest.mark.asyncio
    async def test_json_similarity_partial_match(self) -> None:
        """Test JSON similarity with partial matches."""
        execution = AgentExecution(
            agent_input={"input": "Test"},
            agent_output={"name": "John", "age": 30, "city": "LA"},
            agent_trace=[],
        )
        config = {
            "name": "JsonSimilarityTest",
            "default_evaluation_criteria": {"expected_output": "test"},
        }
        evaluator = JsonSimilarityEvaluator.model_validate(
            {"evaluatorConfig": config, "id": str(uuid.uuid4())}
        )
        criteria = OutputEvaluationCriteria(
            expected_output={"name": "John", "age": 30, "city": "NYC"}  # pyright: ignore[reportCallIssue]
        )

        result = await evaluator.evaluate(execution, criteria)

        assert isinstance(result, NumericEvaluationResult)
        assert math.isclose(result.score, 0.666, abs_tol=1e-3)

    @pytest.mark.asyncio
    async def test_json_similarity_validate_and_evaluate_criteria(self) -> None:
        """Test JSON similarity using validate_and_evaluate_criteria."""
        execution = AgentExecution(
            agent_input={"input": "Test"},
            agent_output={"name": "John", "age": 30, "city": "NYC"},
            agent_trace=[],
        )
        config = {
            "name": "JsonSimilarityTest",
        }
        evaluator = JsonSimilarityEvaluator.model_validate(
            {"evaluatorConfig": config, "id": str(uuid.uuid4())}
        )
        raw_criteria = {"expected_output": {"name": "John", "age": 30, "city": "NYC"}}

        result = await evaluator.validate_and_evaluate_criteria(execution, raw_criteria)

        assert isinstance(result, NumericEvaluationResult)
        assert result.score == 1.0


class TestToolCallOrderEvaluator:
    """Test ToolCallOrderEvaluator.evaluate() method."""

    @pytest.mark.asyncio
    async def test_tool_call_order_perfect_match(
        self, sample_agent_execution_with_trace: AgentExecution
    ) -> None:
        """Test tool call order with perfect order match."""

        config = {
            "name": "ToolOrderTest",
            "strict": True,
        }

        evaluator = ToolCallOrderEvaluator.model_validate(
            {"evaluatorConfig": config, "id": str(uuid.uuid4())}
        )
        criteria = ToolCallOrderEvaluationCriteria(
            tool_calls_order=["tool1", "tool2", "tool1", "tool2"]
        )

        result = await evaluator.evaluate(sample_agent_execution_with_trace, criteria)

        assert isinstance(result, NumericEvaluationResult)
        assert result.score == 1.0

    @pytest.mark.asyncio
    async def test_tool_call_order_no_perfect_match(
        self, sample_agent_execution_with_trace: AgentExecution
    ) -> None:
        """Test tool call order with perfect order match."""

        config = {
            "name": "ToolOrderTest",
            "strict": True,
        }

        evaluator = ToolCallOrderEvaluator.model_validate(
            {"evaluatorConfig": config, "id": str(uuid.uuid4())}
        )
        criteria = ToolCallOrderEvaluationCriteria(
            tool_calls_order=["tool1", "tool1", "tool2", "tool2"]
        )

        result = await evaluator.evaluate(sample_agent_execution_with_trace, criteria)

        assert isinstance(result, NumericEvaluationResult)
        assert result.score == 0.0

    @pytest.mark.asyncio
    async def test_tool_call_order_lcs_match(
        self, sample_agent_execution_with_trace: AgentExecution
    ) -> None:
        """Test tool call order with lcs order match."""

        config = {
            "name": "ToolOrderTest",
            "strict": False,
        }
        evaluator = ToolCallOrderEvaluator.model_validate(
            {"evaluatorConfig": config, "id": str(uuid.uuid4())}
        )
        criteria = ToolCallOrderEvaluationCriteria(
            tool_calls_order=["tool1", "tool1", "tool2", "tool2"]
        )

        result = await evaluator.evaluate(sample_agent_execution_with_trace, criteria)

        assert isinstance(result, NumericEvaluationResult)
        assert result.score == 0.75

    @pytest.mark.asyncio
    async def test_tool_call_order_validate_and_evaluate_criteria(
        self, sample_agent_execution_with_trace: AgentExecution
    ) -> None:
        """Test tool call order using validate_and_evaluate_criteria."""
        config = {
            "name": "ToolOrderTest",
            "strict": True,
        }
        evaluator = ToolCallOrderEvaluator.model_validate(
            {"evaluatorConfig": config, "id": str(uuid.uuid4())}
        )
        raw_criteria = {"tool_calls_order": ["tool1", "tool2", "tool1", "tool2"]}

        result = await evaluator.validate_and_evaluate_criteria(
            sample_agent_execution_with_trace, raw_criteria
        )

        assert isinstance(result, NumericEvaluationResult)
        assert result.score == 1.0


class TestToolCallCountEvaluator:
    """Test ToolCallCountEvaluator.evaluate() method."""

    @pytest.mark.asyncio
    async def test_tool_call_count_exact_match(
        self, sample_agent_execution_with_trace: AgentExecution
    ) -> None:
        """Test tool call count with exact count match."""
        config = {
            "name": "ToolCountTest",
            "strict": True,
        }
        evaluator = ToolCallCountEvaluator.model_validate(
            {"evaluatorConfig": config, "id": str(uuid.uuid4())}
        )
        criteria = ToolCallCountEvaluationCriteria(
            tool_calls_count={"tool1": ("=", 2), "tool2": ("=", 2)}
        )

        result = await evaluator.evaluate(sample_agent_execution_with_trace, criteria)

        assert isinstance(result, NumericEvaluationResult)
        assert result.score == 1.0

    @pytest.mark.asyncio
    async def test_tool_call_count_with_gt(
        self, sample_agent_execution_with_trace: AgentExecution
    ) -> None:
        """Test tool call count with strict count match."""
        config = {
            "name": "ToolCountTest",
            "strict": True,
        }
        evaluator = ToolCallCountEvaluator.model_validate(
            {"evaluatorConfig": config, "id": str(uuid.uuid4())}
        )
        criteria = ToolCallCountEvaluationCriteria(
            tool_calls_count={"tool1": (">", 1), "tool2": (">", 1)}
        )

        result = await evaluator.evaluate(sample_agent_execution_with_trace, criteria)

        assert isinstance(result, NumericEvaluationResult)
        assert result.score == 1.0

    @pytest.mark.asyncio
    async def test_tool_call_count_no_exact_match(
        self, sample_agent_execution_with_trace: AgentExecution
    ) -> None:
        """Test tool call count with no exact count match."""
        config = {
            "name": "ToolCountTest",
            "strict": True,
        }
        evaluator = ToolCallCountEvaluator.model_validate(
            {"evaluatorConfig": config, "id": str(uuid.uuid4())}
        )
        criteria = ToolCallCountEvaluationCriteria(
            tool_calls_count={"tool1": ("=", 2), "tool2": ("=", 1)}
        )

        result = await evaluator.evaluate(sample_agent_execution_with_trace, criteria)

        assert isinstance(result, NumericEvaluationResult)
        assert result.score == 0.0

    @pytest.mark.asyncio
    async def test_tool_call_count_partial_match(
        self, sample_agent_execution_with_trace: AgentExecution
    ) -> None:
        """Test tool call count with partial count match."""
        config = {
            "name": "ToolCountTest",
            "strict": False,
        }
        evaluator = ToolCallCountEvaluator.model_validate(
            {"evaluatorConfig": config, "id": str(uuid.uuid4())}
        )
        criteria = ToolCallCountEvaluationCriteria(
            tool_calls_count={"tool1": ("=", 2), "tool2": ("=", 1)}
        )

        result = await evaluator.evaluate(sample_agent_execution_with_trace, criteria)

        assert isinstance(result, NumericEvaluationResult)
        assert result.score == 0.5

    @pytest.mark.asyncio
    async def test_tool_call_count_validate_and_evaluate_criteria(
        self, sample_agent_execution_with_trace: AgentExecution
    ) -> None:
        """Test tool call count using validate_and_evaluate_criteria."""
        config = {
            "name": "ToolCountTest",
            "strict": True,
        }
        evaluator = ToolCallCountEvaluator.model_validate(
            {"evaluatorConfig": config, "id": str(uuid.uuid4())}
        )
        raw_criteria = {"tool_calls_count": {"tool1": ("=", 2), "tool2": ("=", 2)}}

        result = await evaluator.validate_and_evaluate_criteria(
            sample_agent_execution_with_trace, raw_criteria
        )

        assert isinstance(result, NumericEvaluationResult)
        assert result.score == 1.0


class TestToolCallArgsEvaluator:
    """Test ToolCallArgsEvaluator.evaluate() method."""

    @pytest.mark.asyncio
    async def test_tool_call_args_perfect_match(
        self, sample_agent_execution_with_trace: AgentExecution
    ) -> None:
        """Test tool call args with perfect match."""
        config = {
            "name": "ToolArgsTest",
            "strict": True,
        }
        evaluator = ToolCallArgsEvaluator.model_validate(
            {"evaluatorConfig": config, "id": str(uuid.uuid4())}
        )
        criteria = ToolCallArgsEvaluationCriteria(
            tool_calls=[
                ToolCall(name="tool1", args={"arg1": "value1"}),
                ToolCall(name="tool2", args={"arg2": "value2"}),
                ToolCall(name="tool1", args={"arg1": "value1"}),
                ToolCall(name="tool2", args={"arg2": "value2"}),
            ]
        )

        result = await evaluator.evaluate(sample_agent_execution_with_trace, criteria)

        assert isinstance(result, NumericEvaluationResult)
        assert result.score == 1.0

    @pytest.mark.asyncio
    async def test_tool_call_args_partial_match(
        self, sample_agent_execution_with_trace: AgentExecution
    ) -> None:
        """Test tool call args with partial match."""
        config = {
            "name": "ToolArgsTest",
            "strict": False,
        }
        evaluator = ToolCallArgsEvaluator.model_validate(
            {"evaluatorConfig": config, "id": str(uuid.uuid4())}
        )
        criteria = ToolCallArgsEvaluationCriteria(
            tool_calls=[
                ToolCall(name="tool1", args={"arg1": "value1"}),
                ToolCall(name="tool2", args={"arg2": "value1"}),
                ToolCall(name="tool1", args={"arg1": "value1"}),
                ToolCall(name="tool2", args={"arg2": "value2"}),
            ]
        )

        result = await evaluator.evaluate(sample_agent_execution_with_trace, criteria)

        assert isinstance(result, NumericEvaluationResult)
        assert result.score == 0.75

    @pytest.mark.asyncio
    async def test_tool_call_args_validate_and_evaluate_criteria(
        self, sample_agent_execution_with_trace: AgentExecution
    ) -> None:
        """Test tool call args using validate_and_evaluate_criteria."""
        config = {
            "name": "ToolArgsTest",
            "strict": True,
        }
        evaluator = ToolCallArgsEvaluator.model_validate(
            {"evaluatorConfig": config, "id": str(uuid.uuid4())}
        )
        raw_criteria = {
            "tool_calls": [
                {"name": "tool1", "args": {"arg1": "value1"}},
                {"name": "tool2", "args": {"arg2": "value2"}},
                {"name": "tool1", "args": {"arg1": "value1"}},
                {"name": "tool2", "args": {"arg2": "value2"}},
            ]
        }

        result = await evaluator.validate_and_evaluate_criteria(
            sample_agent_execution_with_trace, raw_criteria
        )

        assert isinstance(result, NumericEvaluationResult)
        assert result.score == 1.0


class TestToolCallOutputEvaluator:
    """Test ToolCallOutputEvaluator.evaluate() method."""

    @pytest.mark.asyncio
    async def test_tool_call_output_perfect_match(
        self, sample_agent_execution_with_trace: AgentExecution
    ) -> None:
        """Test tool call output with perfect output match."""
        config = {
            "name": "ToolOutputTest",
            "strict": True,
        }
        evaluator = ToolCallOutputEvaluator.model_validate(
            {"evaluatorConfig": config, "id": str(uuid.uuid4())}
        )
        criteria = ToolCallOutputEvaluationCriteria(
            tool_outputs=[
                ToolOutput(name="tool1", output="output1"),
                ToolOutput(name="tool2", output="output2"),
                ToolOutput(name="tool1", output="output1"),
                ToolOutput(name="tool2", output="output2"),
            ]
        )

        result = await evaluator.evaluate(sample_agent_execution_with_trace, criteria)

        assert isinstance(result, NumericEvaluationResult)
        assert result.score == 1.0

    @pytest.mark.asyncio
    async def test_tool_call_output_partial_match(
        self, sample_agent_execution_with_trace: AgentExecution
    ) -> None:
        """Test tool call output with partial output match."""
        config = {
            "name": "ToolOutputTest",
            "strict": False,
        }
        evaluator = ToolCallOutputEvaluator.model_validate(
            {"evaluatorConfig": config, "id": str(uuid.uuid4())}
        )
        criteria = ToolCallOutputEvaluationCriteria(
            tool_outputs=[
                ToolOutput(name="tool1", output="output1"),
                ToolOutput(name="tool2", output="wrong_output"),
                ToolOutput(name="tool1", output="output1"),
                ToolOutput(name="tool2", output="output2"),
            ]
        )

        result = await evaluator.evaluate(sample_agent_execution_with_trace, criteria)

        assert isinstance(result, NumericEvaluationResult)
        assert result.score == 0.75

    @pytest.mark.asyncio
    async def test_tool_call_output_no_match_strict(
        self, sample_agent_execution_with_trace: AgentExecution
    ) -> None:
        """Test tool call output with no match in strict mode."""
        config = {
            "name": "ToolOutputTest",
            "strict": True,
        }
        evaluator = ToolCallOutputEvaluator.model_validate(
            {"evaluatorConfig": config, "id": str(uuid.uuid4())}
        )
        criteria = ToolCallOutputEvaluationCriteria(
            tool_outputs=[
                ToolOutput(name="tool1", output="wrong_output1"),
                ToolOutput(name="tool2", output="output2"),
                ToolOutput(name="tool1", output="output1"),
                ToolOutput(name="tool2", output="output2"),
            ]
        )

        result = await evaluator.evaluate(sample_agent_execution_with_trace, criteria)

        assert isinstance(result, NumericEvaluationResult)
        assert result.score == 0.0

    @pytest.mark.asyncio
    async def test_tool_call_output_partial_match_non_strict(
        self, sample_agent_execution_with_trace: AgentExecution
    ) -> None:
        """Test tool call output with partial match in non-strict mode."""
        config = {
            "name": "ToolOutputTest",
            "strict": False,
        }
        evaluator = ToolCallOutputEvaluator.model_validate(
            {"evaluatorConfig": config, "id": str(uuid.uuid4())}
        )
        criteria = ToolCallOutputEvaluationCriteria(
            tool_outputs=[
                ToolOutput(name="tool1", output="wrong_output1"),
                ToolOutput(name="tool2", output="output2"),
            ]
        )

        result = await evaluator.evaluate(sample_agent_execution_with_trace, criteria)

        assert isinstance(result, NumericEvaluationResult)
        assert result.score == 0.5

    @pytest.mark.asyncio
    async def test_tool_call_output_empty_criteria(
        self, sample_agent_execution_with_trace: AgentExecution
    ) -> None:
        """Test tool call output with empty criteria."""
        config = {
            "name": "ToolOutputTest",
            "strict": False,
        }
        evaluator = ToolCallOutputEvaluator.model_validate(
            {"evaluatorConfig": config, "id": str(uuid.uuid4())}
        )
        criteria = ToolCallOutputEvaluationCriteria(tool_outputs=[])

        result = await evaluator.evaluate(sample_agent_execution_with_trace, criteria)

        assert isinstance(result, NumericEvaluationResult)
        assert result.score == 0.0

    @pytest.mark.asyncio
    async def test_tool_call_output_validate_and_evaluate_criteria(
        self, sample_agent_execution_with_trace: AgentExecution
    ) -> None:
        """Test tool call output using validate_and_evaluate_criteria."""
        config = {
            "name": "ToolOutputTest",
            "strict": True,
        }
        evaluator = ToolCallOutputEvaluator.model_validate(
            {"evaluatorConfig": config, "id": str(uuid.uuid4())}
        )
        raw_criteria = {
            "tool_outputs": [
                {"name": "tool1", "output": "output1"},
                {"name": "tool2", "output": "output2"},
                {"name": "tool1", "output": "output1"},
                {"name": "tool2", "output": "output2"},
            ]
        }

        result = await evaluator.validate_and_evaluate_criteria(
            sample_agent_execution_with_trace, raw_criteria
        )

        assert isinstance(result, NumericEvaluationResult)
        assert result.score == 1.0


class TestLlmAsAJudgeEvaluator:
    """Test LlmAsAJudgeEvaluator.evaluate() method."""

    @pytest.mark.asyncio
    async def test_llm_judge_basic_evaluation(
        self, sample_agent_execution: AgentExecution, mocker: MockerFixture
    ) -> None:
        """Test LLM as judge basic evaluation functionality with function calling."""
        # Mock the UiPath constructor to avoid authentication
        mock_uipath = mocker.MagicMock()
        mock_llm = mocker.MagicMock()
        mock_uipath.llm = mock_llm

        # Mock the chat completions response with tool call (function calling approach)
        mock_tool_call = mocker.MagicMock()
        mock_tool_call.id = "call_1"
        mock_tool_call.name = "submit_evaluation"
        mock_tool_call.arguments = {
            "score": 80,
            "justification": "Good response that meets criteria",
        }

        mock_response = mocker.MagicMock()
        mock_response.choices = [
            mocker.MagicMock(
                message=mocker.MagicMock(content=None, tool_calls=[mock_tool_call])
            )
        ]

        # Make chat_completions an async method
        async def mock_chat_completions(*args: Any, **kwargs: Any) -> Any:
            return mock_response

        mock_llm.chat_completions = mock_chat_completions

        mocker.patch("uipath.platform.UiPath", return_value=mock_uipath)

        config = {
            "name": "LlmJudgeTest",
            "prompt": "Rate this output: {{ActualOutput}} vs {{ExpectedOutput}}",
            "model": "gpt-4o-2024-08-06",
        }
        evaluator = LLMJudgeOutputEvaluator.model_validate(
            {"evaluatorConfig": config, "id": str(uuid.uuid4())}
        )

        criteria = OutputEvaluationCriteria(expected_output="Expected output")  # pyright: ignore[reportCallIssue]

        result = await evaluator.evaluate(sample_agent_execution, criteria)

        # Verify the result
        assert hasattr(result, "score")
        assert isinstance(result, NumericEvaluationResult), f"Result is {result}"
        assert result.score == 0.8, f"Result score is {result.score}"

    @pytest.mark.asyncio
    async def test_llm_judge_basic_evaluation_with_llm_service(
        self, sample_agent_execution: AgentExecution, mocker: MockerFixture
    ) -> None:
        """Test LLM judge basic evaluation functionality with a custom LLM service and function calling."""
        # Mock tool call for function calling approach
        mock_tool_call = mocker.MagicMock()
        mock_tool_call.id = "call_1"
        mock_tool_call.name = "submit_evaluation"
        mock_tool_call.arguments = {
            "score": 80,
            "justification": "Good response that meets criteria",
        }

        mock_response = mocker.MagicMock()
        mock_response.choices = [
            mocker.MagicMock(
                message=mocker.MagicMock(content=None, tool_calls=[mock_tool_call])
            )
        ]

        # Make chat_completions an async method
        async def mock_chat_completions(*args: Any, **kwargs: Any) -> Any:
            return mock_response

        config = {
            "name": "LlmJudgeTest",
            "prompt": "Rate this output: {{ActualOutput}} vs {{ExpectedOutput}}",
            "model": "gpt-4o-2024-08-06",
        }
        evaluator = LLMJudgeOutputEvaluator.model_validate(
            {
                "evaluatorConfig": config,
                "llm_service": mock_chat_completions,
                "id": str(uuid.uuid4()),
            }
        )

        criteria = OutputEvaluationCriteria(expected_output="Expected output")  # pyright: ignore[reportCallIssue]

        result = await evaluator.evaluate(sample_agent_execution, criteria)

        # Verify the result
        assert hasattr(result, "score")
        assert isinstance(result, NumericEvaluationResult), f"Result is {result}"
        assert result.score == 0.8, f"Result score is {result.score}"

    @pytest.mark.asyncio
    async def test_llm_judge_validate_and_evaluate_criteria(
        self, sample_agent_execution: AgentExecution, mocker: MockerFixture
    ) -> None:
        """Test LLM judge using validate_and_evaluate_criteria with function calling."""
        # Mock the UiPath constructor to avoid authentication
        mock_uipath = mocker.MagicMock()
        mock_llm = mocker.MagicMock()
        mock_uipath.llm = mock_llm

        # Mock tool call for function calling approach
        mock_tool_call = mocker.MagicMock()
        mock_tool_call.id = "call_1"
        mock_tool_call.name = "submit_evaluation"
        mock_tool_call.arguments = {
            "score": 75,
            "justification": "Good response using raw criteria",
        }

        mock_response = mocker.MagicMock()
        mock_response.choices = [
            mocker.MagicMock(
                message=mocker.MagicMock(content=None, tool_calls=[mock_tool_call])
            )
        ]

        # Make chat_completions an async method
        async def mock_chat_completions(*args: Any, **kwargs: Any) -> Any:
            return mock_response

        mock_llm.chat_completions = mock_chat_completions

        # Mock the UiPath import and constructor
        mocker.patch("uipath.platform.UiPath", return_value=mock_uipath)

        config = {
            "name": "LlmJudgeTest",
            "prompt": "Rate this output: {{ActualOutput}} vs {{ExpectedOutput}}",
            "model": "gpt-4",
        }
        evaluator = LLMJudgeOutputEvaluator.model_validate(
            {"evaluatorConfig": config, "id": str(uuid.uuid4())}
        )
        raw_criteria = {"expected_output": "Expected output"}

        result = await evaluator.validate_and_evaluate_criteria(
            sample_agent_execution, raw_criteria
        )

        # Verify the result
        assert hasattr(result, "score")
        assert isinstance(result, NumericEvaluationResult)
        assert result.score == 0.75


class TestLlmJudgeTrajectoryEvaluator:
    """Test LlmJudgeTrajectoryEvaluator.evaluate() method."""

    @pytest.mark.asyncio
    async def test_llm_trajectory_basic_evaluation(
        self, sample_agent_execution: AgentExecution, mocker: MockerFixture
    ) -> None:
        """Test LLM trajectory judge basic evaluation functionality with function calling."""
        # Mock the UiPath constructor to avoid authentication
        mock_uipath = mocker.MagicMock()
        mock_llm = mocker.MagicMock()
        mock_uipath.llm = mock_llm

        # Mock tool call for function calling approach
        mock_tool_call = mocker.MagicMock()
        mock_tool_call.id = "call_1"
        mock_tool_call.name = "submit_evaluation"
        mock_tool_call.arguments = {
            "score": 90,
            "justification": "The agent followed the expected behavior and met the criteria",
        }

        mock_response = mocker.MagicMock()
        mock_response.choices = [
            mocker.MagicMock(
                message=mocker.MagicMock(content=None, tool_calls=[mock_tool_call])
            )
        ]

        # Make chat_completions an async method
        async def mock_chat_completions(*args: Any, **kwargs: Any) -> Any:
            return mock_response

        mock_llm.chat_completions = mock_chat_completions

        # Mock the UiPath import and constructor
        mocker.patch("uipath.platform.UiPath", return_value=mock_uipath)

        config = {
            "name": "LlmTrajectoryTest",
            "prompt": "Evaluate this trajectory: {{AgentRunHistory}} vs {{ExpectedAgentBehavior}} given the following input: {{UserOrSyntheticInput}} instructions: {{SimulationInstructions}}",
            "model": "gpt-4",
        }
        evaluator = LLMJudgeTrajectoryEvaluator.model_validate(
            {"evaluatorConfig": config, "id": str(uuid.uuid4())}
        )

        criteria = TrajectoryEvaluationCriteria(
            expected_agent_behavior="Agent should respond helpfully"
        )

        result = await evaluator.evaluate(sample_agent_execution, criteria)

        # Verify the result
        assert hasattr(result, "score")
        assert isinstance(result, NumericEvaluationResult)
        assert result.score == 0.9

    @pytest.mark.asyncio
    async def test_llm_trajectory_validate_and_evaluate_criteria(
        self, sample_agent_execution: AgentExecution, mocker: MockerFixture
    ) -> None:
        """Test LLM trajectory judge using validate_and_evaluate_criteria with function calling."""
        # Mock the UiPath constructor to avoid authentication
        mock_uipath = mocker.MagicMock()
        mock_llm = mocker.MagicMock()
        mock_uipath.llm = mock_llm

        # Mock tool call for function calling approach
        mock_tool_call = mocker.MagicMock()
        mock_tool_call.id = "call_1"
        mock_tool_call.name = "submit_evaluation"
        mock_tool_call.arguments = {
            "score": 85,
            "justification": "The agent behavior was good using raw criteria",
        }

        mock_response = mocker.MagicMock()
        mock_response.choices = [
            mocker.MagicMock(
                message=mocker.MagicMock(content=None, tool_calls=[mock_tool_call])
            )
        ]

        # Make chat_completions an async method
        async def mock_chat_completions(*args: Any, **kwargs: Any) -> Any:
            return mock_response

        mock_llm.chat_completions = mock_chat_completions

        # Mock the UiPath import and constructor
        mocker.patch("uipath.platform.UiPath", return_value=mock_uipath)

        config = {
            "name": "LlmTrajectoryTest",
            "prompt": "Evaluate this trajectory: {{AgentRunHistory}} vs {{ExpectedAgentBehavior}} given the following input: {{UserOrSyntheticInput}} instructions: {{SimulationInstructions}}",
            "model": "gpt-4",
        }
        evaluator = LLMJudgeTrajectoryEvaluator.model_validate(
            {"evaluatorConfig": config, "id": str(uuid.uuid4())}
        )
        raw_criteria = {"expected_agent_behavior": "Agent should respond helpfully"}

        result = await evaluator.validate_and_evaluate_criteria(
            sample_agent_execution, raw_criteria
        )

        # Verify the result
        assert hasattr(result, "score")
        assert isinstance(result, NumericEvaluationResult)
        assert result.score == 0.85


class TestEvaluatorErrorHandling:
    """Test error handling in evaluators."""

    @pytest.mark.asyncio
    async def test_invalid_criteria_type(self) -> None:
        """Test that evaluators handle invalid criteria types properly."""
        config = {
            "name": "ErrorTest",
            "default_evaluation_criteria": {"expected_output": "test"},
        }
        evaluator = ExactMatchEvaluator.model_validate(
            {"evaluatorConfig": config, "id": str(uuid.uuid4())}
        )

        with pytest.raises(UiPathEvaluationError):
            # Try to validate invalid criteria
            evaluator.validate_evaluation_criteria("invalid_criteria")

    @pytest.mark.asyncio
    async def test_missing_config_fields(self, mocker: MockerFixture) -> None:
        """Test that evaluators properly validate config fields."""
        # Mock the UiPath constructor to avoid authentication
        mock_uipath = mocker.MagicMock()
        mocker.patch("uipath.platform.UiPath", return_value=mock_uipath)

        config = {
            "name": "LLMJudgeEvaluator",
            "default_evaluation_criteria": {},
        }

        with pytest.raises(ValidationError):
            # Missing required field 'model'
            LLMJudgeOutputEvaluator.model_validate(
                {"evaluatorConfig": config, "id": str(uuid.uuid4())}
            )


class TestEvaluationResultTypes:
    """Test that all evaluators return proper result types."""

    @pytest.mark.asyncio
    async def test_evaluators_return_results_with_scores(
        self, sample_agent_execution: AgentExecution
    ) -> None:
        """Test that evaluators return results with scores."""
        config = {
            "name": "Test",
        }
        evaluator = ExactMatchEvaluator.model_validate(
            {"evaluatorConfig": config, "id": str(uuid.uuid4())}
        )
        criteria = OutputEvaluationCriteria(expected_output={"output": "Test output"})  # pyright: ignore[reportCallIssue]

        result = await evaluator.evaluate(sample_agent_execution, criteria)

        assert hasattr(result, "score")
        assert isinstance(result.score, (int, float))


class TestJustificationHandling:
    """Test justification handling in all evaluators."""

    @pytest.mark.asyncio
    async def test_exact_match_evaluator_justification(
        self, sample_agent_execution: AgentExecution
    ) -> None:
        """Test that ExactMatchEvaluator handles None justification correctly."""
        config = {
            "name": "ExactMatchTest",
            "case_sensitive": True,
        }
        evaluator = ExactMatchEvaluator.model_validate(
            {"evaluatorConfig": config, "id": str(uuid.uuid4())}
        )
        criteria = OutputEvaluationCriteria(expected_output={"output": "Test output"})  # pyright: ignore[reportCallIssue]

        result = await evaluator.evaluate(sample_agent_execution, criteria)

        # Should be NumericEvaluationResult with no justification (None)
        assert isinstance(result, NumericEvaluationResult)
        assert result.score == 1.0
        # Justification should be None for non-LLM evaluators
        assert (
            not hasattr(result, "justification")
            or getattr(result, "justification", None) is None
        )

    @pytest.mark.asyncio
    async def test_json_similarity_evaluator_justification(self) -> None:
        """Test that JsonSimilarityEvaluator handles None justification correctly."""
        execution = AgentExecution(
            agent_input={"input": "Test"},
            agent_output={"name": "John", "age": 30, "city": "NYC"},
            agent_trace=[],
        )
        config = {
            "name": "JsonSimilarityTest",
        }
        evaluator = JsonSimilarityEvaluator.model_validate(
            {"evaluatorConfig": config, "id": str(uuid.uuid4())}
        )
        criteria = OutputEvaluationCriteria(
            expected_output={"name": "John", "age": 30, "city": "NYC"}  # pyright: ignore[reportCallIssue]
        )

        result = await evaluator.evaluate(execution, criteria)

        # Should be NumericEvaluationResult with no justification (None)
        assert isinstance(result, NumericEvaluationResult)
        assert result.score == 1.0
        # Justification should be None for non-LLM evaluators
        assert (
            not hasattr(result, "justification")
            or getattr(result, "justification", None) is None
        )

    @pytest.mark.asyncio
    async def test_tool_call_order_evaluator_justification(
        self, sample_agent_execution_with_trace: AgentExecution
    ) -> None:
        """Test that ToolCallOrderEvaluator handles None justification correctly."""
        config = {
            "name": "ToolOrderTest",
            "strict": True,
        }
        evaluator = ToolCallOrderEvaluator.model_validate(
            {"evaluatorConfig": config, "id": str(uuid.uuid4())}
        )
        criteria = ToolCallOrderEvaluationCriteria(
            tool_calls_order=["tool1", "tool2", "tool1", "tool2"]
        )

        result = await evaluator.evaluate(sample_agent_execution_with_trace, criteria)

        # Should be NumericEvaluationResult with no justification (None)
        assert isinstance(result, NumericEvaluationResult)
        assert result.score == 1.0
        # Justification should be None for non-LLM evaluators
        assert (
            not hasattr(result, "justification")
            or getattr(result, "justification", None) is None
        )

    @pytest.mark.asyncio
    async def test_tool_call_count_evaluator_justification(
        self, sample_agent_execution_with_trace: AgentExecution
    ) -> None:
        """Test that ToolCallCountEvaluator handles None justification correctly."""
        config = {
            "name": "ToolCountTest",
            "strict": True,
        }
        evaluator = ToolCallCountEvaluator.model_validate(
            {"evaluatorConfig": config, "id": str(uuid.uuid4())}
        )
        criteria = ToolCallCountEvaluationCriteria(
            tool_calls_count={"tool1": ("=", 2), "tool2": ("=", 2)}
        )

        result = await evaluator.evaluate(sample_agent_execution_with_trace, criteria)

        # Should be NumericEvaluationResult with no justification (None)
        assert isinstance(result, NumericEvaluationResult)
        assert result.score == 1.0
        # Justification should be None for non-LLM evaluators
        assert (
            not hasattr(result, "justification")
            or getattr(result, "justification", None) is None
        )

    @pytest.mark.asyncio
    async def test_tool_call_args_evaluator_justification(
        self, sample_agent_execution_with_trace: AgentExecution
    ) -> None:
        """Test that ToolCallArgsEvaluator handles None justification correctly."""
        config = {
            "name": "ToolArgsTest",
            "strict": True,
        }
        evaluator = ToolCallArgsEvaluator.model_validate(
            {"evaluatorConfig": config, "id": str(uuid.uuid4())}
        )
        criteria = ToolCallArgsEvaluationCriteria(
            tool_calls=[
                ToolCall(name="tool1", args={"arg1": "value1"}),
                ToolCall(name="tool2", args={"arg2": "value2"}),
                ToolCall(name="tool1", args={"arg1": "value1"}),
                ToolCall(name="tool2", args={"arg2": "value2"}),
            ]
        )

        result = await evaluator.evaluate(sample_agent_execution_with_trace, criteria)

        # Should be NumericEvaluationResult with no justification (None)
        assert isinstance(result, NumericEvaluationResult)
        assert result.score == 1.0
        # Justification should be None for non-LLM evaluators
        assert (
            not hasattr(result, "justification")
            or getattr(result, "justification", None) is None
        )

    @pytest.mark.asyncio
    async def test_tool_call_output_evaluator_justification(
        self, sample_agent_execution_with_trace: AgentExecution
    ) -> None:
        """Test that ToolCallOutputEvaluator handles justification correctly."""
        config = {
            "name": "ToolOutputTest",
            "strict": True,
        }
        evaluator = ToolCallOutputEvaluator.model_validate(
            {"evaluatorConfig": config, "id": str(uuid.uuid4())}
        )
        criteria = ToolCallOutputEvaluationCriteria(
            tool_outputs=[
                ToolOutput(name="tool1", output="output1"),
                ToolOutput(name="tool2", output="output2"),
                ToolOutput(name="tool1", output="output1"),
                ToolOutput(name="tool2", output="output2"),
            ]
        )

        result = await evaluator.evaluate(sample_agent_execution_with_trace, criteria)

        # Should have justification with tool call output details
        assert isinstance(result, NumericEvaluationResult)
        assert result.score == 1.0
        # The justification is stored in the details field for tool call evaluators
        assert hasattr(result, "details")
        assert isinstance(result.details, ToolCallOutputEvaluatorJustification)
        assert hasattr(result.details, "explained_tool_calls_outputs")
        assert isinstance(result.details.explained_tool_calls_outputs, dict)

    @pytest.mark.asyncio
    async def test_llm_judge_output_evaluator_justification(
        self, sample_agent_execution: AgentExecution, mocker: MockerFixture
    ) -> None:
        """Test that LLMJudgeOutputEvaluator handles str justification correctly with function calling."""
        # Mock the UiPath constructor to avoid authentication
        mock_uipath = mocker.MagicMock()
        mock_llm = mocker.MagicMock()
        mock_uipath.llm = mock_llm

        # Mock tool call for function calling approach
        mock_tool_call = mocker.MagicMock()
        mock_tool_call.id = "call_1"
        mock_tool_call.name = "submit_evaluation"
        mock_tool_call.arguments = {
            "score": 80,
            "justification": "The response meets most criteria but could be more detailed",
        }

        mock_response = mocker.MagicMock()
        mock_response.choices = [
            mocker.MagicMock(
                message=mocker.MagicMock(content=None, tool_calls=[mock_tool_call])
            )
        ]

        # Make chat_completions an async method
        async def mock_chat_completions(*args: Any, **kwargs: Any) -> Any:
            return mock_response

        mock_llm.chat_completions = mock_chat_completions
        mocker.patch("uipath.platform.UiPath", return_value=mock_uipath)

        config = {
            "name": "LlmJudgeTest",
            "prompt": "Rate this output: {{ActualOutput}} vs {{ExpectedOutput}}",
            "model": "gpt-4o-2024-08-06",
        }
        evaluator = LLMJudgeOutputEvaluator.model_validate(
            {"evaluatorConfig": config, "id": str(uuid.uuid4())}
        )
        criteria = OutputEvaluationCriteria(expected_output="Expected output")  # pyright: ignore[reportCallIssue]

        result = await evaluator.evaluate(sample_agent_execution, criteria)

        # Should have string justification in details field
        assert isinstance(result, NumericEvaluationResult)
        assert result.score == 0.8
        assert hasattr(result, "details")
        # The justification is stored in the details field for LLM evaluators
        assert isinstance(result.details, str)
        assert (
            result.details
            == "The response meets most criteria but could be more detailed"
        )

    @pytest.mark.asyncio
    async def test_llm_judge_trajectory_evaluator_justification(
        self, sample_agent_execution: AgentExecution, mocker: MockerFixture
    ) -> None:
        """Test that LLMJudgeTrajectoryEvaluator handles str justification correctly."""
        # Mock the UiPath constructor to avoid authentication
        mock_uipath = mocker.MagicMock()
        mock_llm = mocker.MagicMock()
        mock_uipath.llm = mock_llm

        # Mock the chat completions response with justification using tool_call format
        mock_response = mocker.MagicMock()
        mock_tool_call = mocker.MagicMock()
        mock_tool_call.id = "call_1"
        mock_tool_call.name = "submit_evaluation"
        mock_tool_call.arguments = {
            "score": 85,
            "justification": "The agent trajectory shows good decision making and follows expected behavior patterns",
        }

        mock_response.choices = [
            mocker.MagicMock(
                message=mocker.MagicMock(content=None, tool_calls=[mock_tool_call])
            )
        ]

        # Make chat_completions an async method
        async def mock_chat_completions(*args: Any, **kwargs: Any) -> Any:
            return mock_response

        mock_llm.chat_completions = mock_chat_completions
        mocker.patch("uipath.platform.UiPath", return_value=mock_uipath)

        config = {
            "name": "LlmTrajectoryTest",
            "prompt": "Evaluate this trajectory: {{AgentRunHistory}} vs {{ExpectedAgentBehavior}}",
            "model": "gpt-4",
        }
        evaluator = LLMJudgeTrajectoryEvaluator.model_validate(
            {"evaluatorConfig": config, "id": str(uuid.uuid4())}
        )
        criteria = TrajectoryEvaluationCriteria(
            expected_agent_behavior="Agent should respond helpfully"
        )

        result = await evaluator.evaluate(sample_agent_execution, criteria)

        # Should have string justification in details field (not justification attribute)
        assert isinstance(result, NumericEvaluationResult)
        assert result.score == 0.85
        assert isinstance(result.details, str)
        assert (
            result.details
            == "The agent trajectory shows good decision making and follows expected behavior patterns"
        )

    def test_justification_validation_edge_cases(self, mocker: MockerFixture) -> None:
        """Test edge cases for justification validation."""
        # Test None type evaluator
        config_dict = {
            "name": "Test",
            "default_evaluation_criteria": {"expected_output": "test"},
        }
        none_evaluator = ExactMatchEvaluator.model_validate(
            {"evaluatorConfig": config_dict, "id": str(uuid.uuid4())}
        )

        # All inputs should return None for None type evaluators
        assert none_evaluator.validate_justification(None) is None
        assert none_evaluator.validate_justification("") is None
        assert none_evaluator.validate_justification("some text") is None
        assert none_evaluator.validate_justification(123) is None
        assert none_evaluator.validate_justification({"key": "value"}) is None

        # Test str type evaluator - need to provide llm_service to avoid authentication
        llm_config_dict = {
            "name": "Test",
            "default_evaluation_criteria": {"expected_output": "test"},
            "model": "gpt-4o-2024-08-06",
        }
        mock_llm_service = mocker.MagicMock()
        str_evaluator = LLMJudgeOutputEvaluator.model_validate(
            {
                "evaluatorConfig": llm_config_dict,
                "llm_service": mock_llm_service,
                "id": str(uuid.uuid4()),
            }
        )

        # Different inputs should be converted to strings
        assert str_evaluator.validate_justification("test") == "test"
        assert str_evaluator.validate_justification("") == ""
        assert str_evaluator.validate_justification(123) == "123"
        assert str_evaluator.validate_justification(True) == "True"
        assert (
            str_evaluator.validate_justification(None) == ""
        )  # None becomes empty string

    def test_justification_type_extraction_all_evaluators(self) -> None:
        """Test that all evaluators have correct justification type extraction."""
        # Different evaluators have different justification types
        assert ExactMatchEvaluator._extract_justification_type() is type(
            None
        )  # No justification
        assert (
            JsonSimilarityEvaluator._extract_justification_type() is str
        )  # String justification

        # Tool call evaluators have their own justification types
        from uipath.eval.evaluators.tool_call_args_evaluator import (
            ToolCallArgsEvaluatorJustification,
        )
        from uipath.eval.evaluators.tool_call_count_evaluator import (
            ToolCallCountEvaluatorJustification,
        )
        from uipath.eval.evaluators.tool_call_order_evaluator import (
            ToolCallOrderEvaluatorJustification,
        )
        from uipath.eval.evaluators.tool_call_output_evaluator import (
            ToolCallOutputEvaluatorJustification,
        )

        assert (
            ToolCallOrderEvaluator._extract_justification_type()
            is ToolCallOrderEvaluatorJustification
        )
        assert (
            ToolCallCountEvaluator._extract_justification_type()
            is ToolCallCountEvaluatorJustification
        )
        assert (
            ToolCallArgsEvaluator._extract_justification_type()
            is ToolCallArgsEvaluatorJustification
        )
        assert (
            ToolCallOutputEvaluator._extract_justification_type()
            is ToolCallOutputEvaluatorJustification
        )

        # LLM evaluators should have str justification type
        assert LLMJudgeOutputEvaluator._extract_justification_type() is str
        assert LLMJudgeTrajectoryEvaluator._extract_justification_type() is str

    @pytest.mark.asyncio
    async def test_llm_judge_omits_max_tokens_when_none(
        self, sample_agent_execution: AgentExecution, mocker: MockerFixture
    ) -> None:
        """Test that max_tokens is omitted from API request when None (fixes 400 error)."""
        # Mock the UiPath constructor to avoid authentication
        mock_uipath = mocker.MagicMock()
        mock_llm = mocker.MagicMock()
        mock_uipath.llm = mock_llm

        # Mock tool call response
        mock_tool_call = mocker.MagicMock()
        mock_tool_call.id = "call_1"
        mock_tool_call.name = "submit_evaluation"
        mock_tool_call.arguments = {
            "score": 80,
            "justification": "Good response",
        }

        mock_message = mocker.MagicMock()
        mock_message.tool_calls = [mock_tool_call]
        mock_message.content = None

        mock_choice = mocker.MagicMock()
        mock_choice.message = mock_message

        mock_response = mocker.MagicMock()
        mock_response.choices = [mock_choice]

        # Capture the request data passed to chat_completions
        captured_request = {}

        async def capture_chat_completions(**kwargs: Any) -> Any:
            nonlocal captured_request
            captured_request = kwargs
            return mock_response

        mock_llm.chat_completions = capture_chat_completions

        # Patch UiPath to return our mock
        with mocker.patch("uipath.platform.UiPath", return_value=mock_uipath):
            # Create evaluator with max_tokens=None (the default)
            config = {
                "name": "TestMaxTokensNone",
                "model": "gpt-4o-mini-2024-07-18",
                "prompt": "Evaluate the output",
                # max_tokens is intentionally omitted (defaults to None)
            }
            evaluator = LLMJudgeOutputEvaluator.model_validate(
                {"evaluatorConfig": config, "id": str(uuid.uuid4())}
            )

            # Evaluate
            result = await evaluator.evaluate(
                agent_execution=sample_agent_execution,
                evaluation_criteria=OutputEvaluationCriteria(expected_output="42"),
            )

            # Verify max_tokens was NOT included in the request
            assert "max_tokens" not in captured_request, (
                "max_tokens should be omitted when None, not passed as None "
                "(this was causing 400 errors from LLM Gateway API)"
            )

            # Verify other expected fields ARE included
            assert "model" in captured_request
            assert "temperature" in captured_request
            assert "tools" in captured_request
            assert "tool_choice" in captured_request
            assert "messages" in captured_request

            # Verify evaluation result
            assert result.score == 0.8
            assert result.details == "Good response"



from typing import Optional


class UnsupportedDataSourceException(Exception):
    """Raised when an operation is attempted on an unsupported data source type.

    This exception is raised when attempting to use an operation that only supports
    specific data source types (e.g., Orchestrator Storage Bucket) with an incompatible
    data source.
    """

    def __init__(self, operation: str, data_source_type: Optional[str] = None):
        if data_source_type:
            message = f"Operation '{operation}' is not supported for data source type: {data_source_type}. Only Orchestrator Storage Bucket data sources are supported."
        else:
            message = f"Operation '{operation}' requires an Orchestrator Storage Bucket data source."
        super().__init__(message)



"""Unit tests for TYPE_REGISTRY module.

Tests cover:
- get_type() basic functionality
- get_type() error handling
- register_type() custom type registration
- is_valid_type() validation
- Security: ensuring eval() is not used
"""

import pytest

from uipath._cli._utils._type_registry import (
    TYPE_REGISTRY,
    get_type,
    is_valid_type,
    register_type,
)


class TestGetType:
    """Test get_type() function."""

    def test_get_type_str(self):
        """Test getting str type."""
        result = get_type("str")
        assert result is str

    def test_get_type_int(self):
        """Test getting int type."""
        result = get_type("int")
        assert result is int

    def test_get_type_bool(self):
        """Test getting bool type."""
        result = get_type("bool")
        assert result is bool

    def test_get_type_float(self):
        """Test getting float type."""
        result = get_type("float")
        assert result is float

    def test_get_type_invalid_raises_value_error(self):
        """Test that invalid type name raises ValueError."""
        with pytest.raises(ValueError) as exc_info:
            get_type("InvalidType")

        error_msg = str(exc_info.value)
        assert "Unknown type: 'InvalidType'" in error_msg
        assert "Valid types:" in error_msg
        assert "str" in error_msg
        assert "int" in error_msg

    def test_get_type_case_sensitive(self):
        """Test that type names are case-sensitive."""
        with pytest.raises(ValueError):
            get_type("STR")  # Should be "str"

        with pytest.raises(ValueError):
            get_type("Int")  # Should be "int"


class TestIsValidType:
    """Test is_valid_type() function."""

    def test_is_valid_type_str(self):
        """Test that 'str' is valid."""
        assert is_valid_type("str") is True

    def test_is_valid_type_int(self):
        """Test that 'int' is valid."""
        assert is_valid_type("int") is True

    def test_is_valid_type_bool(self):
        """Test that 'bool' is valid."""
        assert is_valid_type("bool") is True

    def test_is_valid_type_float(self):
        """Test that 'float' is valid."""
        assert is_valid_type("float") is True

    def test_is_valid_type_invalid(self):
        """Test that invalid type returns False."""
        assert is_valid_type("InvalidType") is False
        assert is_valid_type("list") is False  # Not registered
        assert is_valid_type("dict") is False  # Not registered

    def test_is_valid_type_empty_string(self):
        """Test that empty string is invalid."""
        assert is_valid_type("") is False


class TestRegisterType:
    """Test register_type() function."""

    def test_register_type_custom(self):
        """Test registering a custom type."""
        from pathlib import Path

        # Clean up if already registered (from previous test run)
        if "path" in TYPE_REGISTRY:
            del TYPE_REGISTRY["path"]

        register_type("path", Path)

        # Verify it's registered
        assert is_valid_type("path")
        assert get_type("path") is Path

        # Clean up
        del TYPE_REGISTRY["path"]

    def test_register_type_duplicate_raises(self):
        """Test that registering duplicate type raises ValueError."""
        with pytest.raises(ValueError) as exc_info:
            register_type("str", str)  # str is already registered

        error_msg = str(exc_info.value)
        assert "str" in error_msg
        assert "already registered" in error_msg

    def test_register_type_persists(self):
        """Test that registered type persists."""
        # Clean up if exists
        if "custom_type" in TYPE_REGISTRY:
            del TYPE_REGISTRY["custom_type"]

        class CustomType:
            pass

        register_type("custom_type", CustomType)

        # Verify persistence
        assert is_valid_type("custom_type")
        assert get_type("custom_type") is CustomType

        # Clean up
        del TYPE_REGISTRY["custom_type"]


class TestTypeRegistrySecurity:
    """Test security aspects of TYPE_REGISTRY."""

    def test_type_registry_does_not_use_eval(self):
        """Test that get_type() does not use eval().

        This is a critical security test ensuring we don't execute
        arbitrary code via eval().
        """
        # Attempt to inject code via type name
        malicious_input = "__import__('os').system('echo pwned')"

        with pytest.raises(ValueError):
            get_type(malicious_input)

        # Verify the malicious code was not executed (would not be in registry)
        assert malicious_input not in TYPE_REGISTRY

    def test_type_registry_only_allows_registered_types(self):
        """Test that only explicitly registered types are allowed."""
        # These types exist in Python but are not in our registry
        unsafe_types = ["list", "dict", "tuple", "set", "object", "type"]

        for unsafe_type in unsafe_types:
            assert not is_valid_type(unsafe_type), (
                f"Type '{unsafe_type}' should not be valid"
            )

            with pytest.raises(ValueError):
                get_type(unsafe_type)


class TestTypeRegistryDefaults:
    """Test default registered types."""

    def test_type_registry_has_expected_defaults(self):
        """Test that TYPE_REGISTRY has all expected default types."""
        expected_types = {"str", "int", "bool", "float"}
        actual_types = set(TYPE_REGISTRY.keys())

        # Check that at least the expected types are present
        # (there might be more if custom types were registered)
        assert expected_types.issubset(actual_types), (
            f"Missing expected types: {expected_types - actual_types}"
        )

    def test_type_registry_str_maps_to_str_class(self):
        """Test that 'str' maps to actual str class."""
        assert TYPE_REGISTRY["str"] is str

    def test_type_registry_int_maps_to_int_class(self):
        """Test that 'int' maps to actual int class."""
        assert TYPE_REGISTRY["int"] is int

    def test_type_registry_bool_maps_to_bool_class(self):
        """Test that 'bool' maps to actual bool class."""
        assert TYPE_REGISTRY["bool"] is bool

    def test_type_registry_float_maps_to_float_class(self):
        """Test that 'float' maps to actual float class."""
        assert TYPE_REGISTRY["float"] is float



"""Tests for JsonSimilarityEvaluator.

Covers exact matches, numeric tolerance, string similarity, and nested structures.
"""

import json
from typing import Any

import pytest

from uipath.eval.evaluators import LegacyJsonSimilarityEvaluator
from uipath.eval.evaluators.base_legacy_evaluator import LegacyEvaluationCriteria
from uipath.eval.models.models import (
    AgentExecution,
    LegacyEvaluatorCategory,
    LegacyEvaluatorType,
)


def _make_base_params() -> dict[str, Any]:
    """Create base parameters for faithfulness evaluator."""
    return {
        "id": "json-sim",
        "category": LegacyEvaluatorCategory.Deterministic,
        "type": LegacyEvaluatorType.JsonSimilarity,
        "name": "JSON Similarity",
        "description": "Compares JSON structures",
        "createdAt": "2025-01-01T00:00:00Z",
        "updatedAt": "2025-01-01T00:00:00Z",
        "targetOutputKey": "*",
    }


class TestJsonSimilarityEvaluator:
    @pytest.mark.asyncio
    async def test_json_similarity_exact_score_1(self) -> None:
        evaluator = LegacyJsonSimilarityEvaluator(**_make_base_params())
        expected_json = """
            {
                "user": {
                    "name": "Alice",
                    "age": 30,
                    "address": {
                        "city": "New York",
                        "zip": "10001"
                    }
                },
                "active": true
            }
            """

        actual_json = """
            {
                "user": {
                    "name": "Alicia",
                    "age": 28,
                    "address": {
                        "city": "New York",
                        "zip": "10002"
                    }
                },
                "active": false,
                "extraField": "Ignored"
            }
            """

        result = await evaluator.evaluate(
            AgentExecution(
                agent_input={},
                agent_trace=[],
                agent_output=json.loads(actual_json),
            ),
            evaluation_criteria=LegacyEvaluationCriteria(
                expected_output=json.loads(expected_json),
                expected_agent_behavior="",
            ),
        )

        assert result.score == 68.0

    @pytest.mark.asyncio
    async def test_json_similarity_exact_score_2(self) -> None:
        evaluator = LegacyJsonSimilarityEvaluator(**_make_base_params())
        expected_json = """
        {
            "users": [
                { "name": "Alice", "age": 25 },
                { "name": "Bob", "age": 30 }
            ]
        }
        """

        actual_json = """
        {
            "users": [
                { "name": "Alice", "age": 24 },
                { "name": "Robert", "age": 30 }
            ],
            "extraField": "Ignored"
        }
        """

        result = await evaluator.evaluate(
            AgentExecution(
                agent_input={},
                agent_trace=[],
                agent_output=json.loads(actual_json),
            ),
            evaluation_criteria=LegacyEvaluationCriteria(
                expected_output=json.loads(expected_json),
                expected_agent_behavior="",
            ),
        )

        assert result.score >= 82.333
        assert result.score <= 82.3334

    @pytest.mark.asyncio
    async def test_json_similarity_exact_score_3(self) -> None:
        evaluator = LegacyJsonSimilarityEvaluator(**_make_base_params())
        expected_json = """
        {
            "name": "Alice",
            "age": 30,
            "active": true
        }
        """

        actual_json = """
        {
            "name": "Alice",
            "age": "30",
            "active": "true"
        }
        """

        result = await evaluator.evaluate(
            AgentExecution(
                agent_input={},
                agent_trace=[],
                agent_output=json.loads(actual_json),
            ),
            evaluation_criteria=LegacyEvaluationCriteria(
                expected_output=json.loads(expected_json),
                expected_agent_behavior="",
            ),
        )

        assert result.score >= 33.333
        assert result.score <= 33.3334

    @pytest.mark.asyncio
    async def test_json_similarity_exact_score_4(self) -> None:
        evaluator = LegacyJsonSimilarityEvaluator(**_make_base_params())
        expected_json = """
        {
          "user": {
            "name": "Alice Johnson",
            "age": 30,
            "active": true,
            "address": {
              "street": "123 Main St",
              "city": "Metropolis",
              "zip": 90210
            }
          },
          "preferences": {
            "newsletter": true,
            "languages": ["en", "fr", "es"]
          },
          "metrics": {
            "visits": 100,
            "ratio": 0,
            "growth": 1.0
          },
          "posts": [
            { "title": "Hello World", "tags": ["intro", "welcome"] },
            { "title": "Deep Dive", "tags": ["advanced", "json", "parsing"] }
          ],
          "permissions": { "admin": false, "editor": true, "viewer": true },
          "notes": null,
          "bio": "Café au lait",
          "scores": [10, 9.5, "N/A", true],
          "settings": {
            "features": [
              { "key": "A", "enabled": true },
              { "key": "B", "enabled": false }
            ]
          }
        }
        """

        actual_json = """
        {
          "user": {
            "name": "Alice Johnson",
            "age": "30",
            "active": "true",
            "address": {
              "street": "123 Main St",
              "city": "Metropolis",
              "zip": "90210"
            }
          },
          "preferences": {
            "newsletter": "yes",
            "languages": ["en", "es", "fr", "de"]
          },
          "metrics": {
            "visits": 98,
            "ratio": 1e-11,
            "growth": 0.9
          },
          "posts": [
            { "title": "Hello-World", "tags": ["intro"] },
            { "title": "Deep  Dive", "tags": ["advanced", "json"] }
          ],
          "permissions": { "admin": "false", "editor": true },
          "bio": "Cafe au lait",
          "scores": [10, "9.5", "NA", "true"],
          "settings": {
            "features": [
              { "key": "B", "enabled": "false" },
              { "key": "A", "enabled": "true" }
            ]
          },
          "extra": { "debug": true }
        }
        """

        result = await evaluator.evaluate(
            AgentExecution(
                agent_input={},
                agent_trace=[],
                agent_output=json.loads(actual_json),
            ),
            evaluation_criteria=LegacyEvaluationCriteria(
                expected_output=json.loads(expected_json),
                expected_agent_behavior="",
            ),
        )

        assert result.score == 43.24977043158861



"""Utility functions for setting evaluation span attributes."""

import json
from typing import Any, Dict, Optional

from opentelemetry.trace import Span, Status, StatusCode
from pydantic import BaseModel, ConfigDict, Field

# Type hint for runtime protocol (avoids circular imports)
try:
    from uipath.runtime import UiPathRuntimeProtocol, UiPathRuntimeSchema
except ImportError:
    UiPathRuntimeProtocol = Any  # type: ignore


class EvalSetRunOutput(BaseModel):
    """Output model for Evaluation Set Run span."""

    model_config = ConfigDict(populate_by_name=True)

    score: int = Field(..., alias="score")


class EvaluationOutput(BaseModel):
    """Output model for Evaluation span."""

    model_config = ConfigDict(populate_by_name=True)

    score: int = Field(..., alias="score")


class EvaluationOutputSpanOutput(BaseModel):
    """Output model for Evaluation output span."""

    model_config = ConfigDict(populate_by_name=True)

    type: int = Field(1, alias="type")
    value: float = Field(..., alias="value")
    evaluator_id: Optional[str] = Field(None, alias="evaluatorId")
    justification: Optional[str] = Field(None, alias="justification")


def calculate_overall_score(evaluator_averages: Dict[str, float]) -> float:
    """Calculate overall average score from evaluator averages.

    Args:
        evaluator_averages: Dictionary mapping evaluator IDs to their average scores

    Returns:
        Overall average score across all evaluators, or 0.0 if no evaluators
    """
    if not evaluator_averages:
        return 0.0
    return sum(evaluator_averages.values()) / len(evaluator_averages)


def calculate_evaluation_average_score(evaluation_run_results: Any) -> float:
    """Calculate average score from evaluation run results.

    Args:
        evaluation_run_results: EvaluationRunResult object containing evaluation results

    Returns:
        Average score across all evaluators, or 0.0 if no results
    """
    if not evaluation_run_results.evaluation_run_results:
        return 0.0

    total_score = sum(
        result.result.score for result in evaluation_run_results.evaluation_run_results
    )
    return total_score / len(evaluation_run_results.evaluation_run_results)


def set_eval_set_run_output_and_metadata(
    span: Span,
    overall_score: float,
    execution_id: str,
    input_schema: Optional[Dict[str, Any]],
    output_schema: Optional[Dict[str, Any]],
    success: bool = True,
) -> None:
    """Set output and metadata attributes for Evaluation Set Run span.

    Args:
        span: The OpenTelemetry span to set attributes on
        overall_score: The overall average score across all evaluators
        execution_id: The execution ID for the evaluation set run
        input_schema: The input schema from the runtime
        output_schema: The output schema from the runtime
        success: Whether the evaluation set run was successful
    """
    # Set span output with overall score using Pydantic model (formatted for UI rendering)
    output = EvalSetRunOutput(score=int(overall_score))
    span.set_attribute("output", output.model_dump_json(by_alias=True, indent=2))

    # Set metadata attributes
    span.set_attribute("agentId", execution_id)
    span.set_attribute("agentName", "N/A")

    # Set schemas as formatted JSON strings for proper rendering in UI
    if input_schema:
        try:
            span.set_attribute("inputSchema", json.dumps(input_schema, indent=2))
        except (TypeError, ValueError):
            span.set_attribute("inputSchema", "{}")
    else:
        span.set_attribute("inputSchema", "{}")

    if output_schema:
        try:
            span.set_attribute("outputSchema", json.dumps(output_schema, indent=2))
        except (TypeError, ValueError):
            span.set_attribute("outputSchema", "{}")
    else:
        span.set_attribute("outputSchema", "{}")

    # Set span status
    if success:
        span.set_status(Status(StatusCode.OK))


def set_evaluation_output_and_metadata(
    span: Span,
    avg_score: float,
    execution_id: str,
    input_data: Optional[Dict[str, Any]] = None,
    has_error: bool = False,
    error_message: Optional[str] = None,
) -> None:
    """Set output and metadata attributes for Evaluation span.

    Args:
        span: The OpenTelemetry span to set attributes on
        avg_score: The average score for this evaluation across all evaluators
        execution_id: The execution ID for this evaluation
        input_data: The input data for this evaluation
        has_error: Whether the evaluation had an error
        error_message: Optional error message if has_error is True
    """
    # Set span output with average score using Pydantic model (formatted for UI rendering)
    output = EvaluationOutput(score=int(avg_score))
    span.set_attribute("output", output.model_dump_json(by_alias=True, indent=2))

    # Set input data if provided (formatted JSON for UI rendering)
    if input_data is not None:
        try:
            span.set_attribute("input", json.dumps(input_data, indent=2))
        except (TypeError, ValueError):
            span.set_attribute("input", "{}")

    # Set metadata attributes
    span.set_attribute("agentId", execution_id)
    span.set_attribute("agentName", "N/A")

    # Set span status based on success
    if has_error and error_message:
        span.set_status(Status(StatusCode.ERROR, error_message))
    elif not has_error:
        span.set_status(Status(StatusCode.OK))


def set_evaluation_output_span_output(
    span: Span,
    score: float,
    evaluator_id: Optional[str] = None,
    justification: Optional[str] = None,
) -> None:
    """Set output attribute for Evaluation output span.

    Args:
        span: The OpenTelemetry span to set attributes on
        score: The evaluation score
        evaluator_id: The ID of the evaluator that produced this score
        justification: Optional justification text for the score
    """
    # Set output using Pydantic model (formatted for UI rendering)
    output = EvaluationOutputSpanOutput(
        value=score,
        evaluator_id=evaluator_id,
        justification=justification,
    )
    span.set_attribute(
        "output", output.model_dump_json(by_alias=True, exclude_none=True, indent=2)
    )


# High-level wrapper functions that handle complete flow


async def configure_eval_set_run_span(
    span: Span,
    evaluator_averages: Dict[str, float],
    execution_id: str,
    schema: UiPathRuntimeSchema,
    success: bool = True,
) -> None:
    """Configure Evaluation Set Run span with output and metadata.

    This high-level function handles:
    - Calculating overall score from evaluator averages
    - Getting runtime schemas
    - Setting all span attributes

    Args:
        span: The OpenTelemetry span to configure
        evaluator_averages: Dictionary mapping evaluator IDs to their average scores
        execution_id: The execution ID for the evaluation set run
        runtime: The runtime instance
        get_schema_func: Async function to get schema from runtime
        success: Whether the evaluation set run was successful
    """
    # Calculate overall score
    overall_score = calculate_overall_score(evaluator_averages)

    # Get runtime schemas
    try:
        input_schema = schema.input
        output_schema = schema.output
    except Exception:
        input_schema = None
        output_schema = None

    # Set span output and metadata
    set_eval_set_run_output_and_metadata(
        span=span,
        overall_score=overall_score,
        execution_id=execution_id,
        input_schema=input_schema,
        output_schema=output_schema,
        success=success,
    )


async def configure_evaluation_span(
    span: Span,
    evaluation_run_results: Any,
    execution_id: str,
    input_data: Optional[Dict[str, Any]] = None,
    agent_execution_output: Optional[Any] = None,
) -> None:
    """Configure Evaluation span with output and metadata.

    This high-level function handles:
    - Calculating average score from evaluation results
    - Determining error status
    - Setting all span attributes

    Args:
        span: The OpenTelemetry span to configure
        evaluation_run_results: EvaluationRunResult object containing evaluation results
        execution_id: The execution ID for this evaluation
        input_data: The input data for this evaluation
        agent_execution_output: Optional agent execution output for error checking
    """
    # Calculate average score
    avg_score = calculate_evaluation_average_score(evaluation_run_results)

    # Determine error status
    has_error = False
    error_message = None
    if agent_execution_output is not None:
        try:
            if agent_execution_output.result.error:
                has_error = True
                error_message = str(agent_execution_output.result.error)
        except (AttributeError, NameError, UnboundLocalError):
            pass

    # Set span output and metadata
    set_evaluation_output_and_metadata(
        span=span,
        avg_score=avg_score,
        execution_id=execution_id,
        input_data=input_data,
        has_error=has_error,
        error_message=error_message,
    )



import logging
from typing import Optional

logger: logging.Logger = logging.getLogger("uipath")


def setup_logging(should_debug: Optional[bool] = None) -> None:
    """Configure logging for the CLI."""
    if not logger.handlers:
        handler = logging.StreamHandler()
        handler.setFormatter(
            logging.Formatter(
                fmt="%(message)s",
                datefmt="%Y-%m-%d %H:%M:%S",
            )
        )
        logger.addHandler(handler)
        logger.setLevel(logging.DEBUG if should_debug else logging.INFO)

        # Prevent propagation to root logger to avoid duplicate logs
        logger.propagate = False



"""LLM request throttling utilities.

This module provides concurrency control for LLM API requests to prevent
overwhelming the system with simultaneous calls.
"""

import asyncio

DEFAULT_LLM_CONCURRENCY = 20
_llm_concurrency_limit: int = DEFAULT_LLM_CONCURRENCY
_llm_semaphore: asyncio.Semaphore | None = None
_llm_semaphore_loop: asyncio.AbstractEventLoop | None = None


def get_llm_semaphore() -> asyncio.Semaphore:
    """Get the LLM semaphore, creating with configured limit if not set.

    The semaphore is recreated if called from a different event loop than
    the one it was originally created in. This prevents "bound to a different
    event loop" errors when using multiple asyncio.run() calls.
    """
    global _llm_semaphore, _llm_semaphore_loop

    loop = asyncio.get_running_loop()

    # Recreate semaphore if it doesn't exist or if the event loop changed
    if _llm_semaphore is None or _llm_semaphore_loop is not loop:
        _llm_semaphore = asyncio.Semaphore(_llm_concurrency_limit)
        _llm_semaphore_loop = loop

    return _llm_semaphore


def set_llm_concurrency(limit: int) -> None:
    """Set the max concurrent LLM requests. Call before making any LLM calls.

    Args:
        limit: Maximum number of concurrent LLM requests allowed (must be > 0).

    Raises:
        ValueError: If limit is less than 1.
    """
    if limit < 1:
        raise ValueError("LLM concurrency limit must be at least 1")

    global _llm_concurrency_limit, _llm_semaphore, _llm_semaphore_loop
    _llm_concurrency_limit = limit
    _llm_semaphore = None
    _llm_semaphore_loop = None



"""UiPath Orchestrator Models.

This module contains models related to UiPath Orchestrator services.
"""

from ._assets_service import AssetsService
from ._attachments_service import AttachmentsService
from ._buckets_service import BucketsService
from ._folder_service import FolderService
from ._jobs_service import JobsService
from ._mcp_service import McpService
from ._processes_service import ProcessesService
from ._queues_service import QueuesService
from .assets import Asset, UserAsset
from .attachment import Attachment
from .buckets import Bucket, BucketFile
from .job import Job, JobErrorInfo
from .mcp import McpServer, McpServerStatus, McpServerType
from .processes import Process
from .queues import (
    CommitType,
    QueueItem,
    QueueItemPriority,
    TransactionItem,
    TransactionItemResult,
)

__all__ = [
    "AssetsService",
    "AttachmentsService",
    "BucketsService",
    "FolderService",
    "JobsService",
    "McpService",
    "ProcessesService",
    "QueuesService",
    "Asset",
    "UserAsset",
    "Attachment",
    "Bucket",
    "BucketFile",
    "Job",
    "JobErrorInfo",
    "Process",
    "CommitType",
    "QueueItem",
    "QueueItemPriority",
    "TransactionItem",
    "TransactionItemResult",
    "McpServer",
    "McpServerStatus",
    "McpServerType",
]
